#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from hemisphere_plot_utility import get_euler_projection, get_top_projection, plot_hemisphere, plot_axis
from screenshot_utility import Screenshot, Magnifier
from polygon_clipping import clip_polygon
from polygon_sampling_ported import prepare_projected_solid_angle_polygon_sampling, \
    sample_projected_solid_angle_polygon, sample_sector_between_ellipses, is_inner_ellipse, \
    get_ellipse_direction_factor, solve_homogeneous_quadratic, rotate_90, mix_fma, \
    prepare_solid_angle_polygon_sampling, sample_solid_angle_polygon
from polygon_sampling_related_work_ported import prepare_projected_solid_angle_polygon_sampling_arvo, \
    sample_projected_solid_angle_polygon_arvo
from matplotlib import pyplot, cm, colors
from imageio import imwrite, imread
import re
import subprocess
from os import listdir, path

# A few colors that are used consistently
light_color = np.asarray([1.0, 0.9, 0.5])
light_outline_color = np.asarray([0.95, 0.8, 0.4])
solid_angle_color = np.asarray([1.0, 0.9, 0.5])
solid_angle_outline_color = np.asarray(3 * [0.0])
solid_angle_line_color = np.asarray([0.95, 0.8, 0.4])
horizon_color = np.asarray(3 * [0.85 * 0.9])
great_circle_color = np.asarray([0.3, 0.6, 0.3])
sector_color = np.asarray(3 * [0.6])
vertex_color = np.asarray(3 * [0.0])
sampled_area_color = np.asarray([0.3, 0.75, 1.0])
sampled_area_line_color = np.asarray([0.25, 0.65, 0.9])
sampled_area_outline_color = np.asarray([0.2, 0.5, 0.9])
current_area_color = np.asarray([0.6, 1.0, 0.4])
area_discrepancy_color = np.asarray([1.0, 0.6, 0.3])

# The width of a column in the acmtog article layout in centimeters
column_width = 8.5579


def get_figure_size(width_in_centimeters, height_in_centimeters=None, scaling=1.5):
    """
    Returns a figure size as expected by matplotlib.pyplot.figure. It is
    computed from the given size in centimeters. If no Height or Width is given,
    the aspect ratio is fixed to be 4:3. The figure size is scaled up by the
    given amount.
    """
    if not height_in_centimeters:
        height_in_centimeters = width_in_centimeters * 0.75
    if not width_in_centimeters:
        width_in_centimeters = height_in_centimeters / 0.75
    return width_in_centimeters * scaling / 2.54, height_in_centimeters * scaling / 2.54


def srgb_to_linear(srgb):
    """
    Converts all scalars in the given array (which should be normalized to the
    range from zero to one) from an sRGB scale to linear scale.
    """
    srgb = np.maximum(0.0, srgb)
    return np.where(srgb <= 0.04045, srgb / 12.92, ((srgb + 0.055) / 1.055) ** 2.4)

def linear_to_srgb(linear):
    """
    Converts all scalars in the given array (which should be normalized to the
    range from zero to one) from a linear scale to the sRGB scale.
    """
    linear = np.maximum(0.0, linear)
    return np.where(linear <= 0.0031308, 12.92 * linear, 1.055 * linear ** (1.0/2.4) - 0.055)


def list_experiments(file_name_re, experiment_path="../vulkan_renderer/data/experiments"):
    """
    Goes through the given directory and lists files matching a given regular
    expression.
    :param file_name_re: A regular expression for file names.
    :param experiment_path: The directory that is searched.
    :return: A list of pairs (file_path, name). file_path is the complete path
        to a file with matching file name in the given subdirectory (not
        recursively). name is the file name with timing information between an
        underscore and the file extension removed.
    """
    return [(path.join(experiment_path, file_name), re.sub(r"^(.*?)_([0-9]+spp_)?[0-9.]+(\.\w+)$", r"\1\3", file_name))
            for file_name in listdir(experiment_path)
            if re.match(file_name_re, file_name) is not None]


def compute_error(image, reference, metric="rmse", return_image=False):
    """
    Given two HDR images where values of 1.0 correspond to saturated pixels,
    this function evaluates an error metric that is zero for identical images.
    :param image: An image with error.
    :param reference: An image without (or at least with little) error.
    :param metric: Can be "rmse" for the RMSE on unclamped images,
        "clamped_rmse" for the RMSE on images clamped to one (per channel) or
        "rrmse" for relative RMSE.
    :param return_image: Pass True to get an additional return value.
    :return: A scalar error. If return_image is True, the function additionally
        returns an image with large entries in regions that contribute a lot to
        the scalar error.
    """
    # Set NaN or INF pixels to zero
    mask = (image[:, :, 0] == 1.0) & (image[:, :, 1] == 0.0) & (image[:, :, 2] == 0.796875)
    for j in range(image.shape[2]):
        image[:, :, j] = np.where(mask, 0.0, image[:, :, j])
    if metric == "rmse":
        difference = image - reference
    elif metric == "clamped_rmse":
        difference = np.minimum(image, 1.0) - np.minimum(reference, 1.0)
    elif metric == "rrmse":
        difference = (image - reference) / np.where(reference == 0.0, 1.0, reference)
        difference = np.where(reference == 0.0, 0.0, difference)
    elif metric == "robust_rrmse":
        reference_luminance = 0.2126 * reference[:, :, 0] + 0.7152 * reference[:, :, 1] + 0.0722 * reference[:, :, 2]
        difference = (image - reference) / (reference_luminance[:, :, np.newaxis] + 0.1)
    else:
        print("Unknown metric:", metric)
        return
    mean_factor = 1.0 / np.sqrt(reference.shape[0] * reference.shape[1])
    error = np.linalg.norm(difference) * mean_factor
    if return_image:
        return error, difference
    else:
        return error


def print_experiment_errors(file_name_re, reference_name, metric, plot_error=False, experiment_path="../vulkan_renderer/data/experiments"):
    """
    Lists experiments through list_experiments(). The specified files should be
    *.hdr. It then computes errors with respect to one of the experiments and
    prints them.
    :param file_name_re: Forwards to list_experiments().
    :param reference_name: The name of the reference image as returned by
        list_experiments().
    :param metric: See compute_error().
    :param plot_error: Each error metric computes the two-norm of some form of
        difference image. Pass True to open a figure of these images.
    :param experiment_path: Forwards to list_experiments().
    """
    experiments = list_experiments(file_name_re, experiment_path)
    experiments = sorted(experiments)
    references = [path for path, name in experiments if name == reference_name]
    if len(references) != 1:
        print("There are %d references of name %s among experiments matching %s. Need exactly one." % (len(references), reference_name, file_name_re))
        return
    reference = imread(references[0])
    print("Printing %s with respect to %s:" % (metric, reference_name))
    plot_index = 1
    for path, name in experiments:
        print(name, end=":\t")
        image = imread(path)
        error, difference = compute_error(image, reference, metric, True)
        print("%.3f" % error)
        if plot_error:
            pyplot.subplot(3, (len(experiments) + 2) // 3, plot_index, title="%s, %f" % (name, error), xticks=[], yticks=[])
            plot_index += 1
            pyplot.imshow(np.linalg.norm(difference, axis=2), vmin=0.0, vmax=4.0)
    if plot_error:
        pyplot.show()


def prepare_polygon(polygon):
    """
    Prepares a polygon for further processing and plotting.
    :param polygon: An array of shape (vertex_count, 3) providing all vertices.
    :return: Three representations of the polygon as arrays of shape
        (vertex_count, 3) where vertex_count may be different for the clipped
        versions. The result is a triple (polygon, clipped, projected) where
        polygon is the original polygon, clipped is clipped against the plane
        z=0 and projected is clipped with vertices on the unit sphere.
    """
    polygon = np.asarray(polygon)
    clipped_vertex_count, clipped_polygon = clip_polygon(np.copy(polygon), polygon.shape[0] + 1)
    clipped_polygon = clipped_polygon[:clipped_vertex_count, :]
    projected_polygon = clipped_polygon / np.linalg.norm(clipped_polygon, axis=-1)[:, np.newaxis]
    return polygon, clipped_polygon, projected_polygon


def get_decentral_quad():
    """
    Returns a rectangle that yields a decentral case for sampling. The output
    format is the same as for prepare_polygon().
    """
    # Define the polygon (a rectangle)
    origin = np.asarray([1.2, 1.3, 0.4])
    direction_u = np.asarray([-2.3, -0.5, -1.2])
    direction_v = np.asarray([-0.4, 0.4, 3.3])
    direction_v -= (direction_u @ direction_v) / (direction_u @ direction_u) * direction_u
    polygon = np.asarray([origin, origin + direction_u, origin + direction_u + direction_v, origin + direction_v])
    return prepare_polygon(polygon)


def get_central_quad():
    """
    Returns a rectangle that yields a central case for sampling. The output
    format is the same as for prepare_polygon().
    """
    # Define the polygon (a rectangle)
    origin = np.asarray([1.3, -1.2, 0.8])
    direction_u = np.asarray([0.5, 2.3, 0.4])
    direction_v = np.asarray([-1.9, 0.4, 0.3])
    direction_v -= (direction_u @ direction_v) / (direction_u @ direction_u) * direction_u
    polygon = np.asarray([origin, origin + direction_u, origin + direction_u + direction_v, origin + direction_v])
    return prepare_polygon(polygon)


def get_decentral_zenith_polygon():
    """
    Returns a polygon that yields a decentral case for sampling with edges
    through the zenith. The output format is the same as for prepare_polygon().
    """
    # Define the polygon (a rectangle)
    origin = np.asarray([1.2, 1.3, 1.3])
    direction_u = np.asarray([-3.3, -0.5, 0.0])
    direction_v = np.asarray([0.0, 0.0, 2.0])
    direction_v -= (direction_u @ direction_v) / (direction_u @ direction_u) * direction_u
    polygon = np.asarray([origin, origin + direction_u, origin + direction_u + direction_v, origin + direction_v])
    polygon = np.vstack([polygon[:1, :], [[-0.1, 1.0, 0.7]], polygon[1:, :]])
    return prepare_polygon(polygon)


def get_central_zenith_quad():
    """
    Returns a rectangle that yields a central case for sampling with one edge
    through the zenith. The output format is the same as for prepare_polygon().
    """
    # Define the polygon (a rectangle)
    origin = np.asarray([1.3, -0.247807, 0.8])
    direction_u = np.asarray([0.5, 2.3, 0.4])
    direction_v = np.asarray([-1.9, 0.4, 0.3])
    direction_v -= (direction_u @ direction_v) / (direction_u @ direction_u) * direction_u
    polygon = np.asarray([origin, origin + direction_u, origin + direction_u + direction_v, origin + direction_v])
    return prepare_polygon(polygon)


def get_initialization_quad():
    """
    Returns a rectangle that poses issues for a simple initialization strategu
    but not for the one we use. The output format is the same as for
    prepare_polygon().
    """
    origin = np.asarray([1.7, 0.1, 0.0])
    direction_u = np.asarray([-3.4, 0.0, 0.0])
    direction_v = np.asarray([0.0, 0.3, 1.5])
    direction_v -= (direction_u @ direction_v) / (direction_u @ direction_u) * direction_u
    polygon = np.asarray([origin, origin + direction_u, origin + direction_u + direction_v, origin + direction_v])
    return prepare_polygon(polygon)


def prepare_sampling(prepared_polygon):
    """
    Invokes prepare_projected_solid_angle_polygon_sampling() for the given
    output of prepare_polygon() and returns the resulting
    projected_solid_angle_polygon_t.
    """
    clipped = prepared_polygon[1]
    clipped = np.vstack([clipped, clipped[0:1, :]])
    return prepare_projected_solid_angle_polygon_sampling(clipped.shape[0] - 1, clipped)


def sample_edges(vertices, close_loop):
    """
    This method is useful for plotting the solid angle of a polygon or other
    arcs of great circles.
    :param vertices: Array of shape (vertex_count, 3) providing vertices of the
        edges. Any two subsequent vertices get connected by an edge. Need not be
        normalized.
    :param close_loop: Pass True to ensure that the first and last vertex are
        connected by an edge.
    :return: An array of shape (sample_count, 3) providing samples of all
        specified edges as points on the unit sphere.
    """
    vertices = np.asarray(vertices)
    lerp = np.linspace(0.0, 1.0, 513)[:, np.newaxis]
    if close_loop:
        vertices = np.vstack([vertices, vertices[0:1, :]])
    vertices /= np.linalg.norm(vertices, axis=1)[:, np.newaxis]
    edges = [(1.0 - lerp) * vertices[j, np.newaxis, :] + lerp * vertices[j + 1, np.newaxis, :] for j in range(vertices.shape[0] - 1)]
    result = np.concatenate(edges)
    result /= np.linalg.norm(result, axis=-1)[:, np.newaxis]
    return result


def plot_polygon(prepared_polygon, projection, elements=("hemisphere", "polygon", "clipped_polygon", "projected_polygon", "great_circles", "edge_normals", "ellipse_normals", "vertices", "sectors")):
    """
    Creates a plot showing the given polygon with selected elements.
    :param prepared_polygon: Output of prepare_polygon().
    :param projection: The used OrthogonalProjection.
    :param elements: A sequence of elements that should be included in the
        plot. It consists of any combination of the following strings:
        - hemisphere: A hemisphere around the shading point in the origin
        - horizon: The great circle for the horizon
        - polygon: The untransformed and unclipped polygon
        - polygon_red: The untransformed and unclipped polygon colored red
        - clipped_polygon: The untransformed clipped polygon
        - projected_polygon: The clipped polygon projected onto the unit sphere
        - projected_polygon_no_fill: Same but without a fill
        - projected_polygon_thin: Same but without a fill and a thin outline
        - great_circles: Great circles through each edge of the clipped polygon
        - great_circle_%d: Only the great circle with the given index
        - edge_normals: Normal vectors for all edges with visible great circles
        - ellipse_normals: Edge normals extended to the plane z=1
        - unlabeled_normals: ellipse_normals without edge
        - simple_label_normals: edge_normals without division by z in the label
        - vertices: Vertices projected to the unit sphere with labels
        - vertex_lines: Subtle lines connecting the origin and clipped vertices
        - sectors: The sectors in the xy-plane that are relevant for sampling
    """
    polygon, clipped_polygon, projected_polygon = prepared_polygon

    def mask_polygon(points, used_polygon):
        """
        :return: True for points occluded by the given polygon.
        """
        if np.linalg.det(used_polygon[0:3, :]) < 0.0:
            used_polygon = used_polygon[::-1, :]
        depth_axis = projection.matrix[2, :]
        matrices = np.zeros(list(points.shape)[:-1] + [used_polygon.shape[0], 3, 3])
        matrices[..., 0, :] = points[..., np.newaxis, :] - used_polygon
        matrices[..., 1, :] = depth_axis
        matrices[..., 2, :] = np.roll(used_polygon, 1, axis=0) - used_polygon
        overlapped = np.all(np.linalg.det(matrices) < 0.0, axis=-1)
        return overlapped

    def mask_projected_polygon(points):
        """
        :return: True for points in the solid angle of the polygon or occluded
            by it.
        """
        matrices = np.zeros(list(points.shape)[:-1] + [clipped_polygon.shape[0], 3, 3])
        matrices[..., 0, :] = points[..., np.newaxis, :]
        matrices[..., 1, :] = clipped_polygon
        matrices[..., 2, :] = np.roll(clipped_polygon, 1, axis=0)
        in_polygon = np.all(np.linalg.det(matrices) < 0.0, axis=-1)
        # Flip the points along the depth axis and test for inclusion again
        depth_axis = projection.matrix[2, :]
        flipped_points = points - 2.0 * np.einsum("i, j, ...j -> ...i", depth_axis, depth_axis, points)
        matrices[..., 0, :] = flipped_points[..., np.newaxis, :]
        in_flipped_polygon = np.all(np.linalg.det(matrices) < 0.0, axis=-1)
        # We only care about points behind the polygon
        depth = np.einsum("i, ...i -> ...", depth_axis, points)
        behind_polygon = in_flipped_polygon & (depth < 0.0)
        return in_polygon | behind_polygon

    # Plot the various elements
    if "hemisphere" in elements:
        bigger_polygon = None
        if "polygon" in elements or "clipped_polygon" in elements:
            bigger_polygon = polygon if "polygon" in elements else clipped_polygon
        if bigger_polygon is not None and "projected_polygon" not in elements:
            def mask(points): return mask_polygon(points, bigger_polygon)
        elif bigger_polygon is None and "projected_polygon" in elements:
            mask = mask_projected_polygon
        elif bigger_polygon is not None:
            def mask(points): return mask_projected_polygon(points) | mask_polygon(points, bigger_polygon)
        else:
            mask = None
        if mask is None:
            plot_hemisphere(projection)
        else:
            plot_hemisphere(projection, 8, 4, mask, color=solid_angle_line_color)
            plot_hemisphere(projection, 8, 4, lambda points: ~mask(points))
    if "horizon" in elements:
        phase = np.linspace(-np.pi, np.pi, 1025)
        circle = np.zeros((phase.size, 3))
        circle[:, 0] = np.cos(phase)
        circle[:, 1] = np.sin(phase)
        projection.plot_curve(circle, color=horizon_color)
    if "polygon" in elements or "polygon_red" in elements:
        red = "polygon_red" in elements
        screen_polygon = projection.transform(polygon)
        pyplot.fill(screen_polygon[:, 0], screen_polygon[:, 1], zorder=-100.0, color=[1.0, 0.5, 0.3] if red else light_color)
        screen_polygon = np.vstack([screen_polygon, screen_polygon[0:1, :]])
        pyplot.plot(screen_polygon[:, 0], screen_polygon[:, 1], zorder=-90.0, color=[0.9, 0.3, 0.2] if red else light_outline_color)
    if "clipped_polygon" in elements:
        screen_polygon = projection.transform(clipped_polygon)
        pyplot.fill(screen_polygon[:, 0], screen_polygon[:, 1], zorder=-100.0, color=light_color)
        screen_polygon = np.vstack([screen_polygon, screen_polygon[0:1, :]])
        pyplot.plot(screen_polygon[:, 0], screen_polygon[:, 1], zorder=20.0, color=light_outline_color)
    if "projected_polygon" in elements or "projected_polygon_no_fill" in elements or "projected_polygon_thin" in elements:
        # Plot the outline
        outline_curve = sample_edges(clipped_polygon, True)
        if "projected_polygon_thin" in elements:
            projection.plot_curve(outline_curve, color=3 * [0.3], linewidth=1.0, linestyle=":", zorder=10.0)
        else:
            projection.plot_curve(outline_curve, color=solid_angle_outline_color, zorder=10.0)
        # For the area handle front- and backfacing parts separately
        for depth_factor in [-1.0, 1.0][0:(2 if "projected_polygon" in elements else 0)]:
            colored = "projected_polygon" in elements
            # Transform and clip to find the relevant parts
            screen_vertices = projection.transform(clipped_polygon)
            screen_vertices[..., 2] *= depth_factor
            new_vertex_count, screen_vertices = clip_polygon(screen_vertices, screen_vertices.shape[0] + 1)
            if new_vertex_count > 0:
                screen_vertices = screen_vertices[0:new_vertex_count, :]
                outline = sample_edges(screen_vertices, True)
                pyplot.fill(outline[:, 0], outline[:, 1], zorder=-100.0, color=solid_angle_color)

    if "great_circles" in elements:
        great_circle_list = range(clipped_polygon.shape[0])
    else:
        great_circle_list = [int(element[len("great_circle_"):]) for element in elements if element[0:len("great_circle_")] == "great_circle_"]
    phase = np.linspace(-np.pi, np.pi, 1025)
    for edge in great_circle_list:
        vertex_a = projected_polygon[edge, :]
        vertex_b = np.copy(projected_polygon[(edge + 1) % projected_polygon.shape[0], :])
        vertex_b -= (vertex_b @ vertex_a) * vertex_a
        vertex_b /= np.linalg.norm(vertex_b)
        circle = np.cos(phase)[:, np.newaxis] * vertex_a[np.newaxis, :] + np.sin(phase)[:, np.newaxis] * vertex_b[np.newaxis, :]
        projection.plot_curve(circle, color=great_circle_color)
        # Plot edge normals and ellipse normals
        normal = np.cross(clipped_polygon[edge, :], clipped_polygon[(edge + 1) % clipped_polygon.shape[0], :])
        screen_normal = projection.transform(normal)
        screen_normal /= normal[2]
        normal_types = [
            ("edge_normals", screen_normal, r"$\frac{n_%d}{n_{%d,z}}$" % (edge, edge)),
            ("ellipse_normals", screen_normal, "$u_%d$" % edge),
            ("unlabeled_normals", screen_normal, ""),
            ("simple_label_normals", screen_normal, r"$n_%d$" % edge),
        ]
        for element, vector, label in normal_types:
            if element not in elements:
                continue
            pyplot.arrow(0.0, 0.0, vector[0], vector[1], head_width=0.08, edgecolor=great_circle_color, facecolor="w")
            text_offset = 0.11 * vector / np.linalg.norm(vector)
            text_offset = np.asarray([-text_offset[1], text_offset[0]])
            text_location = vector[0:2] * 0.5 * np.minimum(1.0, 1.0 / np.linalg.norm(vector[0:2])) + text_offset
            pyplot.text(text_location[0], text_location[1], label, ha="center", va="center")
    if "vertices" in elements:
        normalized_polygon = clipped_polygon / np.linalg.norm(clipped_polygon, axis=1)[:, np.newaxis]
        screen_polygon = projection.transform(normalized_polygon)
        pyplot.scatter(screen_polygon[:, 0], screen_polygon[:, 1], color=vertex_color, s=4.5, zorder=10000.0)
        for i in range(clipped_polygon.shape[0]):
            left_vector = screen_polygon[i - 1, 0:2] - screen_polygon[i, 0:2]
            right_vector = screen_polygon[(i + 1) % clipped_polygon.shape[0], 0:2] - screen_polygon[i, 0:2]
            half_vector = left_vector / np.linalg.norm(left_vector) + right_vector / np.linalg.norm(right_vector)
            half_vector /= np.linalg.norm(half_vector)
            offset = -0.1 if clipped_polygon[i, 2] <= 0.001 else 0.1
            location = screen_polygon[i, 0:2] - offset * half_vector
            pyplot.text(location[0], location[1], "%d" % i, ha="center", va="center")
    if "vertex_lines" in elements:
        for j in range(clipped_polygon.shape[0]):
            vertex = projection.transform(clipped_polygon[j, :])
            pyplot.plot([0.0, vertex[0]], [0.0, vertex[1]], color=3 * [0.75], linestyle=":", linewidth=0.8, zorder=-100.0)
    if "sectors" in elements:
        for vertex in range(clipped_polygon.shape[0]):
            direction = projected_polygon[vertex, 0:2] / np.linalg.norm(projected_polygon[vertex, 0:2])
            sector = np.asarray([
                [0.0, 0.0, 0.0],
                [direction[0], direction[1], 0.0]])
            projection.plot_curve(sector, color=sector_color)
        # screen_polygon = projection.transform(projected_polygon)
        # pyplot.scatter(screen_polygon[:, 0], screen_polygon[:, 1], color="C1", s=6.0, zorder=20.0)


def plot_geometric_constructions():
    """
    Creates plots to show the geometric constructions that characterize the
    projected solid angle of the polygon through ellipses.
    """
    shared_elements = ["projected_polygon", "hemisphere", "great_circle_4"]
    prepared_polygon = get_decentral_quad()
    # 3D version
    size_ratio_3d = 0.555
    size_ratio_2d = 1.0 - size_ratio_3d
    projection = get_euler_projection(-0.56 * np.pi, 0.23 * np.pi)
    figure = pyplot.figure(figsize=get_figure_size(column_width * size_ratio_3d, column_width * size_ratio_2d))
    pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=(-1.02, 1.34), ylim=(-0.68, 1.2))
    plot_polygon(prepared_polygon, projection, shared_elements + ["polygon_red", "clipped_polygon", "edge_normals", "vertex_lines"])
    for axis in range(3):
        plot_axis(projection, [0.5, -0.57, 0.0], axis, True, 0.09)
    pyplot.tight_layout(pad=0.0, h_pad=0.0, w_pad=0.0)
    figure.savefig("../Graphics/polygon/geometric_constructions/3d.pdf")
    # 2D version
    projection = get_top_projection()
    figure = pyplot.figure(figsize=get_figure_size(column_width * size_ratio_2d, column_width * size_ratio_2d))
    pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=(-1.01, 1.01), ylim=(-1.01, 1.01))
    plot_polygon(prepared_polygon, projection, shared_elements + ["vertices", "ellipse_normals"])
    for axis in range(2):
        plot_axis(projection, [-0.92, -0.92, 0.0], axis, axis == 1, 0.09)
    pyplot.tight_layout(pad=0.0, h_pad=0.0, w_pad=0.0)
    figure.savefig("../Graphics/polygon/geometric_constructions/2d.pdf")


def plot_case_distinction():
    """
    Plots illustrative examples of the central and decentral case.
    """
    projection = get_top_projection()
    plots = [
        ("central", get_central_quad(), 1, 0.45 * np.pi),
        ("decentral", get_decentral_quad(), 2, 0.77 * np.pi),
        ("central_zenith", get_central_zenith_quad(), 1, 0.45 * np.pi),
        ("decentral_zenith", get_decentral_zenith_polygon(), 2, 0.77 * np.pi),
    ]
    for name, prepared_polygon, sector_index, sample_azimuth in plots:
        figure = pyplot.figure(figsize=get_figure_size(column_width * 0.4, column_width * 0.4))
        pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=(-1.01, 1.01), ylim=(-1.01, 1.01))
        plot_polygon(prepared_polygon, projection, ["horizon", "projected_polygon", "vertices", "sectors"])
        # Construct a vertex in direction of the sampled direction
        clipped_polygon = prepared_polygon[1]
        edge_normal = np.cross(clipped_polygon[sector_index, :], clipped_polygon[sector_index + 1, :])
        if edge_normal[2] < 0.0:
            edge_normal = -edge_normal
        new_vertex = np.cross(edge_normal, [np.sin(sample_azimuth), -np.cos(sample_azimuth), 0.0])
        # Shade the area clockwise of the sample
        if name.find("decentral") >=0:
            # Plot the area clockwise of the sampled area
            clipped_polygon = prepared_polygon[1]
            rotation = np.asarray([
                [-np.cos(sample_azimuth), -np.sin(sample_azimuth), 0.0],
                [0.0, 0.0, 1.0],
                [np.sin(sample_azimuth), -np.cos(sample_azimuth), 0.0],
            ]).T
            clipped_vertex_count, sampled_polygon = clip_polygon(clipped_polygon @ rotation, clipped_polygon.shape[0] + 1)
            sampled_polygon = sampled_polygon[0:clipped_vertex_count, :] @ rotation.T
        else:
            sampled_polygon = np.vstack([clipped_polygon[0:sector_index + 1, :], new_vertex, [[0.0, 0.0, 1.0]]])
        outline = sample_edges(sampled_polygon, True)
        screen_outline = projection.transform(outline)
        pyplot.fill(screen_outline[:, 0], screen_outline[:, 1], color=sampled_area_color)
        # Draw a labeled line towards the sampled vertex
        sampled_dir = np.asarray([np.cos(sample_azimuth), np.sin(sample_azimuth)])
        pyplot.plot([0.0, sampled_dir[0]], [0.0, sampled_dir[1]], color=sampled_area_outline_color)
        offset = np.asarray([-sampled_dir[1], sampled_dir[0]])
        text_location = 0.85 * sampled_dir - 0.1 * offset
        pyplot.text(text_location[0], text_location[1], "$w$", ha="center", va="center")
        # Plot x and y axis
        for axis in range(2):
            plot_axis(projection, [-0.92, -0.92, 0.0], axis, axis == 1, 0.09)
        pyplot.tight_layout(pad=0.0, h_pad=0.0, w_pad=0.0)
        figure.savefig("../Graphics/polygon/case_distinction/%s.pdf" % name)


def sample_single_quad(random_numbers, inner_ellipse, outer_ellipse, dir_0, dir_1):
    """
    Samples the quad formed by the intersection points of the given normalized
    direction vectors with the given ellipses. The implementation is a reduced
    version of sample_sector_between_ellipses().
    """
    normalization_factor = np.asarray([
        [
            get_ellipse_direction_factor(inner_ellipse, dir_0),
            get_ellipse_direction_factor(inner_ellipse, dir_1)
        ],
        [
            get_ellipse_direction_factor(outer_ellipse, dir_0),
            get_ellipse_direction_factor(outer_ellipse, dir_1)
        ]
    ]).T
    sector_area = normalization_factor[0][1] * normalization_factor[1][1] - normalization_factor[0][0] * normalization_factor[1][0]
    target_quad_area = -sector_area * random_numbers[0]
    target_quad_area *= np.linalg.det(np.vstack([dir_0, dir_1]))
    quad_normals = [
        dir_0 * normalization_factor[0][0] + dir_1 * normalization_factor[1][0],
        dir_0 * normalization_factor[0][1] + dir_1 * normalization_factor[1][1]
    ]
    quad_normals[0] += (inner_ellipse @ quad_normals[0]) * inner_ellipse
    quad_normals[1] += (outer_ellipse @ quad_normals[1]) * outer_ellipse
    quad_offsets = [
        (quad_normals[0] @ dir_0) * normalization_factor[1][0],
        (quad_normals[1] @ dir_0) * normalization_factor[1][1]
    ]
    quadratic = np.outer((quad_offsets[1] * normalization_factor[0][1]) * rotate_90(dir_1), quad_normals[0])
    quadratic -= np.outer((quad_offsets[0] * normalization_factor[0][0]) * rotate_90(dir_1) + target_quad_area * quad_normals[0], quad_normals[1])
    current_dir = solve_homogeneous_quadratic(quadratic)
    current_dir *= np.sign(np.linalg.det(np.vstack([dir_0, current_dir])))
    inner_factor = get_ellipse_direction_factor(inner_ellipse, current_dir) ** 2
    outer_factor = get_ellipse_direction_factor(outer_ellipse, current_dir) ** 2
    current_dir *= np.sqrt(mix_fma(inner_factor, outer_factor, random_numbers[1]))
    return current_dir


def plot_polygon_sampling(prepared_polygon, elements, iteration, sector, random_numbers, initialization_random_number=None, ellipse_label_angle=0.095 * np.pi):
    """
    Plots constructions for sampling of ellipses within a sector.
    :param prepared_polygon: Output of prepare_polygon().
    :param elements: A sequence of strings defining the elements to include in
        the plot. Can be any combination of:
        - sector: The boundaries of the selected sector
        - sector_half: The half-vector of the selected sector
        - ellipses: The ellipses between which the sampling takes place
        - area: The area to be sampled (as filled polygon)
        - horizon: The unit circle in the xy-plane
        - single_quad: A single quad formed from intersections of sector
          boundaries with ellipses
        - init_quads: The pair of quads used for the initialization
        - current_direction: The direction in the current iteration
        - next_direction: The direction after the current iteration
        - current_area: The area resulting from the current iteration
        - tangents: The tangents that are used for an iteration
        - tangent_points: The points where the tangents meet the ellipses
        - tangent_area: The area enclosed between the tangents for an update
        - tangent_area_red: Same with a different color
        - sampled_point: The sampled point
        - samples_%d: Uniform samples after the given number of iterations
        - single_quad_samples: Samples produced by a single-quad initialization
    :param iteration: How many iterations should be performed to define the
        current direction.
    :param sector: The index of the sector that should be sampled
    :param random_numbers: The random numbers used to sample the azimuth and
        the radial component within the sector
    :param initialization_random_number: If you want to force a bad
        initialization, you can pass a different random number for it.
    :param ellipse_label_angle: Ellipse labels are placed at this angle relative
        to the sector end.
    """
    # If we want to plot sampled points, we use these uniform blue noise points
    random_number_array = np.asarray([[0.306725, 0.43533], [0.824745, 0.033991], [0.9392575, 0.875415], [0.6764, 0.76945], [0.4415775, 0.2484225], [0.8640025, 0.12221275], [0.22939975, 0.6516075], [0.43499, 0.88808], [0.2785625, 0.7928175], [0.877755, 0.415695], [0.6467975, 0.5180525], [0.70811, 0.241596], [0.627685, 0.5949575], [0.190045, 0.2671375], [0.7515925, 0.5356475], [0.4863125, 0.3181775], [0.10410625, 0.7692175], [0.6101625, 0.816915], [0.608635, 0.7214475], [0.5622175, 0.5445375], [0.511755, 0.21709725], [0.664155, 0.675655], [0.2238825, 0.57773], [0.2637525, 0.08962075], [0.0625925, 0.99672], [0.832095, 0.5717025], [0.412965, 0.7215825], [0.4003, 0.96349], [0.793345, 0.2540375], [0.6402275, 0.20258375], [0.6535475, 0.9704925], [0.736015, 0.70215], [0.831535, 0.3584775], [0.36281, 0.5639125], [0.5879575, 0.8995275], [0.4093675, 0.1877765], [0.9632175, 0.5166825], [0.4791175, 0.401105], [0.566795, 0.1220785], [0.148027, 0.84561], [0.036093, 0.206373], [0.3101325, 0.696255], [0.5095575, 0.488185], [0.7737125, 0.4133], [0.35017, 0.763655], [0.3373025, 0.849645], [0.2631175, 0.8990775], [0.9443475, 0.2601075], [0.5531225, 0.03688125], [0.43574, 0.556955], [0.332375, 0.199918], [0.3429975, 0.2759775], [0.301375, 0.6113825], [0.06652825, 0.5408175], [0.121947, 0.219138], [0.24631075, 0.3739375], [0.18350525, 0.085252], [0.51567, 0.9122275], [0.39587, 0.04911575], [0.930685, 0.186746], [0.90285, 0.5690975], [0.4982525, 0.14325], [0.85803, 0.209564], [0.14717975, 0.9183475], [0.6391225, 0.1240445], [0.331605, 0.359465], [0.7209425, 0.84635], [0.7699, 0.617085], [0.5513275, 0.3902725], [0.220677, 0.8420125], [0.7000075, 0.60006], [0.0246693, 0.01775387], [0.863685, 0.8080275], [0.17596375, 0.41168], [0.01704612, 0.7708], [0.21392125, 0.95994], [0.408195, 0.31868], [0.71174, 0.15619875], [0.2877975, 0.5232275], [0.2597075, 0.22835225], [0.2614475, 0.3045275], [0.784705, 0.17557975], [0.482445, 0.06105325], [0.439065, 0.473555], [0.1049645, 0.2889175], [0.5142925, 0.8380225], [0.229103, 0.01095323], [0.068666, 0.62702], [0.628425, 0.3037175], [0.233741, 0.725505], [0.5003975, 0.5903375], [0.1204365, 0.4646875], [0.8826375, 0.4900925], [0.7242725, 0.930365], [0.9699525, 0.03144125], [0.163292, 0.1588955], [0.176568, 0.7609475], [0.786015, 0.098255], [0.6576775, 0.88215], [0.8982475, 0.039498], [0.9238925, 0.7260125], [0.9621325, 0.44148], [0.23779675, 0.16307175], [0.6449275, 0.0520615], [0.626885, 0.3968225], [0.344545, 0.9217325], [0.06974125, 0.8426875], [0.04627825, 0.4689075], [0.940725, 0.797875], [0.29264, 0.9824], [0.420475, 0.1172995], [0.556065, 0.30694], [0.07481425, 0.0740345], [0.13572875, 0.580545], [0.909095, 0.952675], [0.034597, 0.2989225], [0.5176025, 0.690915], [0.3994725, 0.3990175], [0.5751325, 0.9851775], [0.3338675, 0.1236605], [0.944945, 0.6318025], [0.838435, 0.98379], [0.812415, 0.6764], [0.878845, 0.6647825], [0.946025, 0.1132105], [0.13646075, 0.99572], [0.7760225, 0.9732325], [0.36519, 0.482505], [0.5502675, 0.76659], [0.8724525, 0.2935125], [0.155261, 0.34425], [0.3162075, 0.03524275], [0.0847995, 0.144482], [0.00339535, 0.6997475], [0.82526, 0.75095], [0.6983475, 0.404125], [0.47278, 0.76713], [0.05838325, 0.9135175], [0.7645925, 0.3213675], [0.9928525, 0.9528325], [0.5705525, 0.63898], [0.0862505, 0.6994825], [0.71913, 0.0658075], [0.01058852, 0.3858525], [0.0818855, 0.38228], [0.5799825, 0.22744175], [0.376395, 0.6467475], [0.23248225, 0.46522], [0.70468, 0.476845], [0.6931625, 0.32399], [0.757605, 0.7835275], [0.48672, 0.9750375], [0.806855, 0.4800975], [0.9294025, 0.3539025], [0.17710975, 0.517325], [0.0025264, 0.11470875], [0.5880875, 0.463975], [0.4513325, 0.65081], [0.136729, 0.03072325], [0.413245, 0.818605], [0.858215, 0.891435], [0.7950675, 0.8690425], [0.154159, 0.6626975]])
    # Prepare the polygon for the plotting operations
    prepared_sampling = prepare_sampling(prepared_polygon)
    # Select a sector
    inner_ellipse = prepared_sampling.inner_ellipse_0
    for j in range(sector + 1):
        vertex_ellipse = prepared_sampling.ellipses[j]
        if j == 0:
            outer_ellipse = vertex_ellipse
        else:
            inner_ellipse = vertex_ellipse if is_inner_ellipse(vertex_ellipse) else inner_ellipse
            outer_ellipse = outer_ellipse if is_inner_ellipse(vertex_ellipse) else vertex_ellipse
    dirs = prepared_sampling.vertices[sector:sector + 2]
    for j in range(2):
        dirs[j] = dirs[j] / np.linalg.norm(dirs[j])
    ellipses = [inner_ellipse, outer_ellipse]
    ellipse_sqrts = list()
    for ellipse in ellipses:
        square_root = 1.0 / np.sqrt(np.sqrt(ellipse @ ellipse + 1.0) + 1.0) * ellipse
        ellipse_sqrts.append(np.eye(2) + np.outer(square_root, square_root))
    # Perform the requested number of iterations
    target_area = random_numbers[0] * prepared_sampling.sector_projected_solid_angles[sector]
    used_random_numbers = random_numbers if initialization_random_number is None else np.asarray([initialization_random_number, random_numbers[1]])
    current_dir = sample_sector_between_ellipses(used_random_numbers, target_area, inner_ellipse, outer_ellipse, dirs[0], dirs[1], iteration)
    next_dir = sample_sector_between_ellipses(used_random_numbers, target_area, inner_ellipse, outer_ellipse, dirs[0], dirs[1], iteration + 1)
    # Sample a circle for a few plots
    phase = np.linspace(-np.pi, np.pi, 1025)
    circle = np.vstack([np.cos(phase), np.sin(phase)]).T
    # Construct tangent lines and ompute where the current and next direction
    # intersect them
    rotate_90 = np.asarray([[0.0, -1.0], [1.0, 0.0]])
    tangent_lines = list()
    tangent_edges = list()
    tangent_points = list()
    for ellipse in ellipses:
        ellipse_matrix = np.eye(2) + np.outer(ellipse, ellipse)
        tangent_normal = current_dir @ ellipse_matrix
        tangent_dir = tangent_normal @ rotate_90
        tangent_dir *= 100.0 / np.linalg.norm(tangent_dir)
        normalization = 1.0 / np.sqrt((current_dir @ current_dir) + (current_dir @ ellipse) ** 2)
        ellipse_point = normalization * current_dir
        tangent_points.append(ellipse_point)
        tangent_lines.append([(ellipse_point[j] - tangent_dir[j], ellipse_point[j] + tangent_dir[j]) for j in range(2)])
        tangent_edge = list()
        for edge_dir in [current_dir, next_dir]:
            edge_dir *= (tangent_normal @ ellipse_point) / (tangent_normal @ edge_dir)
            tangent_edge.append(edge_dir)
        tangent_edges.append(np.asarray(tangent_edge))

    def shade_sector(dirs, color):
        """Shades the area between the ellipses in the given sector."""
        for i in range(2):
            dirs[i] = dirs[i] / np.linalg.norm(dirs[i])
        orthogonal = dirs[1] - (dirs[1] @ dirs[0]) * dirs[0]
        orthogonal /= np.linalg.norm(orthogonal)
        sector_angle = np.arccos(dirs[0] @ dirs[1])
        phase = np.linspace(0.0, sector_angle, 257)
        arc = np.cos(phase)[:, np.newaxis] * dirs[0][np.newaxis, :] + np.sin(phase)[:, np.newaxis] * orthogonal[np.newaxis, :]
        ellipse_arcs = list()
        for ellipse in ellipses:
            normalization = 1.0 / np.sqrt(1.0 + (arc @ ellipse) ** 2)
            ellipse_arcs.append(normalization[:, np.newaxis] * arc)
        outline = np.concatenate([ellipse_arcs[0], ellipse_arcs[1][::-1]], axis=0)
        pyplot.fill(outline[:, 0], outline[:, 1], color=color, zorder=-99.0)

    # Plot the various elements
    display_dirs = []
    if "sector" in elements:
        display_dirs += [("$s_0$", -0.06, 0.3, dirs[0]), ("$s_1$", 0.06, 0.3, dirs[1])]
    if "sector_half" in elements:
        display_dirs += [("$s_h$", 0.06, 0.15, (dirs[0] + dirs[1]) / np.linalg.norm(dirs[0] + dirs[1]))]
    for label, offset, distance, direction in display_dirs:
        pyplot.plot([0.0, direction[0]], [0.0, direction[1]], color=sector_color)
        text_location = distance * direction
        text_location += offset * rotate_90 @ direction
        pyplot.text(text_location[0], text_location[1], label, ha="center", va="center")
    if "ellipses" in elements:
        for ellipse_sqrt in ellipse_sqrts:
            curve = circle @ np.linalg.inv(ellipse_sqrt)
            pyplot.plot(curve[:, 0], curve[:, 1], color=great_circle_color)
        label_rotation = ellipse_label_angle
        label_dir = np.asarray([[np.cos(label_rotation), -np.sin(label_rotation)], [np.sin(label_rotation), np.cos(label_rotation)]]) @ dirs[1]
        for label, ellipse, offset in [("$C_i$", inner_ellipse, -0.08), ("$C_o$", outer_ellipse, 0.08)]:
            ellipse_matrix = np.eye(2) + np.outer(ellipse, ellipse)
            text_position = label_dir / np.sqrt(np.dot(label_dir, ellipse_matrix @ label_dir))
            normal = ellipse_matrix @ label_dir
            normal /= np.linalg.norm(normal)
            text_position += offset * normal
            pyplot.text(text_position[0], text_position[1], label, ha="center", va="center")
    if "area" in elements:
        shade_sector(dirs, solid_angle_color)
    if "horizon" in elements:
        pyplot.plot(circle[:, 0], circle[:, 1], color=horizon_color)
    for element in ["single_quad", "init_quads"]:
        if element not in elements:
            continue
        if element == "single_quad":
            all_dirs = [dirs[0], dirs[1]]
        else:
            all_dirs = [dirs[0], (dirs[0] + dirs[1]) / np.linalg.norm(dirs[0] + dirs[1]), dirs[1]]
        for j in range(2):
            quad_dirs = all_dirs[j:j + 2]
            quad_edges = list()
            for ellipse in ellipses:
                normalization = 1.0 / np.sqrt(1.0 + (np.asarray(quad_dirs) @ ellipse) ** 2)
                quad_edges.append(normalization[:, np.newaxis] * quad_dirs)
            outline = np.concatenate([quad_edges[0], quad_edges[1][::-1]], axis=0)
            pyplot.fill(outline[:, 0], outline[:, 1], color=sampled_area_color, zorder=-98.0)
            outline = np.vstack([outline, outline[0:1, :]])
            pyplot.plot(outline[:, 0], outline[:, 1], color=sampled_area_outline_color)
    label_side = np.sign(np.linalg.det(tangent_edges[1]))
    if "current_direction" in elements:
        pyplot.plot([0.0, tangent_edges[1][0, 0]], [0.0, tangent_edges[1][0, 1]], color=sector_color)
        text_location = (tangent_edges[0][0, :] + tangent_edges[1][0, :]) * 0.5
        text_location[0] += -label_side * 0.02
        pyplot.text(text_location[0], text_location[1], "$w_n$", ha=("left" if label_side < 0.0 else "right"), va="center")
    if "next_direction" in elements:
        pyplot.plot([0.0, tangent_edges[1][1, 0]], [0.0, tangent_edges[1][1, 1]], color=sector_color)
        text_location = (tangent_edges[0][1, :] + tangent_edges[1][1, :]) * 0.5
        text_location[0] -= -label_side * 0.01
        pyplot.text(text_location[0], text_location[1], "$w_{n+1}$", ha=("right" if label_side < 0.0 else "left"), va="center")
    if "current_area" in elements:
        shade_sector([dirs[0], current_dir], current_area_color)
    if "tangent_area" in elements or "tangent_area_red" in elements:
        outline = np.concatenate([tangent_edges[0], tangent_edges[1][::-1]], axis=0)
        pyplot.fill(outline[:, 0], outline[:, 1], color=(sampled_area_color if "tangent_area" in elements else area_discrepancy_color))
        text_location = outline.mean(axis=0)
        pyplot.text(text_location[0], text_location[1], "$A_\delta$", ha="center", va="center")
    if "tangents" in elements:
        pyplot.plot(*tangent_lines[0], color="k")
        pyplot.plot(*tangent_lines[1], color="k")
    if "tangent_points" in elements:
        pyplot.scatter(tangent_points[0][0], tangent_points[0][1], c="w", edgecolors="k", s=22.0, zorder=300.0)
        pyplot.scatter(tangent_points[1][0], tangent_points[1][1], c="w", edgecolors="k", s=22.0, zorder=300.0)
    if "sampled_point" in elements:
        pyplot.scatter([current_dir[0]], [current_dir[1]], color="C2", zorder=200.0)
    sample_iteration_counts = [int(element[len("samples_"):]) for element in elements if element[:len("samples_")] == "samples_"]
    for sample_iteration_count in sample_iteration_counts:
        samples = list()
        for sample_random_numbers in random_number_array:
            samples.append(sample_sector_between_ellipses(sample_random_numbers, sample_random_numbers[0] * prepared_sampling.sector_projected_solid_angles[sector], inner_ellipse, outer_ellipse, dirs[0], dirs[1], sample_iteration_count))
        samples = np.asarray(samples)
        pyplot.scatter(samples[:, 0], samples[:, 1], color="C1", s=1.5, zorder=100.0)
    if "single_quad_samples" in elements:
        samples = list()
        for sample_random_numbers in random_number_array:
            samples.append(sample_single_quad(sample_random_numbers, inner_ellipse, outer_ellipse, dirs[0], dirs[1]))
        samples = np.asarray(samples)
        pyplot.scatter(samples[:, 0], samples[:, 1], color="C1", s=1.5, zorder=100.0)


def plot_iteration():
    """
    Plots the iterative sampling procedure.
    """
    width = 2.8
    figure = pyplot.figure(figsize=get_figure_size(column_width, column_width * 1.04 / width))
    pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=(-0.5 * width, 0.5 * width), ylim=(-0.02, 1.02))
    plot_polygon(get_decentral_quad(), get_top_projection(), ["projected_polygon_thin"])
    elements = ("sector", "ellipses", "area", "current_direction", "next_direction", "tangents", "tangent_area")
    plot_polygon_sampling(get_decentral_quad(), elements, 0, 2, np.asarray([0.8, 0.4]), 0.52)
    pyplot.tight_layout(pad=0.0, h_pad=0.0, w_pad=0.0)
    figure.savefig("../Graphics/polygon/iteration/iteration.pdf")


def plot_initizialization():
    """
    Plots the initialization of our sampling procedure and an alternative with a
    single quad.
    """
    for plot_elements in [["single_quad", "single_quad_samples"], ["init_quads", "samples_0", "sector_half"]]:
        figure = pyplot.figure(figsize=get_figure_size(column_width * 0.5, column_width * 0.26))
        pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=(-1.02, 1.02), ylim=(-0.03, 1.01))
        elements = ["sector", "ellipses", "area"] + plot_elements
        plot_polygon_sampling(get_initialization_quad(), elements, 0, 1, np.asarray([0.8, 0.4]), None, -0.1 * np.pi)
        pyplot.tight_layout(pad=0.0, h_pad=0.0, w_pad=0.0)
        figure.savefig("../Graphics/polygon/initialization/%s.pdf" % plot_elements[0])


def plot_arvo_failure():
    """
    Visualizes failure cases of Arvo's method for sampling the projected solid
    angle of polygons.
    """
    point_data_file = "blue_noise_de_goes.dat"
    random_number_array = np.fromfile(point_data_file, sep=" ").reshape((-1, 2))
    permutation = np.argsort(random_number_array[:, 0])
    random_number_array[:, 0] = random_number_array[:, 0][permutation]
    random_number_array[:, 1] = random_number_array[:, 1][permutation]
    random_number_array = random_number_array[:, :]
    # A triangle that provokes issues in the initialization
    initialization_polygon = np.asarray([
        [-0.001, 1.0, 0.0],
        [0.0, 0.8, 1.0],
        [ 1.0, 0.25, 0.0],
    ])
    rotation = 0.12 * np.pi
    initialization_polygon = initialization_polygon @ np.asarray([
        [np.cos(rotation), np.sin(rotation), 0.0],
        [-np.sin(rotation), np.cos(rotation), 0.0],
        [0.0, 0.0, 1.0],
    ])
    # A triangle that provokes divergence
    iteration_polygon = np.asarray([
        [-1.0, 0.25, 0.0],
        [-0.8, 0.35, 1.0],
        [ 1.0, 0.25, 0.0],
    ])
    plots = [
        (0, initialization_polygon, 0, (-0.38, 0.82), (0.57, 1.01), (0.0, 0.39), 1.2),
        (1, iteration_polygon, 11, (-1.0, 1.0), (0.227, 1.01), (0.0, 0.0), 0.72)
    ]
    figure_width = 0.5 * column_width
    figure = pyplot.figure(figsize=get_figure_size(figure_width, figure_width * 0.7))
    axes_list = list()
    for i, polygon, iteration_count, xlim, ylim, pos, scale in plots:
        # Prepare the polygon for sampling
        prepared_polygon = prepare_polygon(polygon)
        projection = get_top_projection()
        padded_polygon = np.vstack([prepared_polygon[2], prepared_polygon[2]])
        prepared_sampling_arvo = prepare_projected_solid_angle_polygon_sampling_arvo(prepared_polygon[2].shape[0], padded_polygon)
        # Produce samples with our method
        prepared_sampling = prepare_projected_solid_angle_polygon_sampling(prepared_polygon[2].shape[0], padded_polygon)
        samples = list()
        for sample_random_numbers in random_number_array:
            samples.append(sample_projected_solid_angle_polygon(prepared_sampling, np.copy(sample_random_numbers)))
        samples = np.asarray(samples)
        width, height = xlim[1] - xlim[0], ylim[1] - ylim[0]
        axes_pos = (pos[0], pos[1], scale * width, scale * height)
        axes = pyplot.axes(axes_pos, aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=xlim, ylim=ylim)
        axes_list.append(axes)
        plot_polygon(prepared_polygon, projection, ["projected_polygon", "horizon", "ellipse_normals"])
        samples_arvo = list()
        for sample_random_numbers in random_number_array:
            samples_arvo.append(sample_projected_solid_angle_polygon_arvo(prepared_sampling_arvo, np.copy(sample_random_numbers), iteration_count))
        samples_arvo = np.asarray(samples_arvo)
        pyplot.plot(samples[:, 0], samples[:, 1], "o", mfc="w", mec="k", fillstyle="full", mew=0.25, ms=1.5, zorder=90.0)
        pyplot.plot(samples_arvo[:, 0], samples_arvo[:, 1], "o", color="C0", fillstyle="full", mew=0.0, ms=1.5, zorder=100.0)
        # lines = np.concatenate([samples[np.newaxis, :, :], samples_arvo[np.newaxis, :, :]], axis=0)
        # pyplot.plot(lines[:, :, 0], lines[:, :, 1], color="k", linewidth=0.2, zorder=100.0)
        label_size = 6.8
        if i == 0:
            pyplot.text(0.42, 0.95, "(a) Initialization", fontsize=label_size)
        else:
            pyplot.text(0.0, 0.34, "(b) 11 iterations", fontsize=label_size)
    figure.savefig("../Graphics/polygon/arvo_failure/combined.pdf", bbox_inches="tight", pad_inches=0.0)


def plot_ltc_space(name, world_to_cosine_space, polygon_world_space):
    """
    Implements plot_ltc() using the given transform to construct the linearly
    transformed cosine and showing the given polygon in world space.
    """
    # Set up a perspective projection
    yaw_angle, pitch_angle = -0.1 * np.pi, 0.03 * np.pi
    projection = get_euler_projection(yaw_angle, pitch_angle)
    # Plot a hemisphere
    figure = pyplot.figure(figsize=get_figure_size(column_width * 0.5, column_width))
    pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False)
    hemisphere_color = np.asarray([80, 78, 134]) / 255.0
    plot_hemisphere(projection, 0, 1, color=hemisphere_color)
    half_phase = np.linspace(-0.5 * np.pi, 0.5 * np.pi, 512)
    pyplot.plot(np.sin(half_phase), np.cos(half_phase), color=hemisphere_color)
    # Plot the horizon
    normal_cosine_space = np.asarray([0.0, 0.0, 1.0])
    normal_world_space = world_to_cosine_space.T @ normal_cosine_space
    horizon_tangent = np.cross(projection.matrix[2, :], normal_world_space)
    horizon_tangent /= np.linalg.norm(horizon_tangent)
    horizon_bitangent = np.cross(horizon_tangent, normal_world_space)
    horizon_bitangent /= np.linalg.norm(horizon_bitangent)
    for sign, style in [(-1.0, "-"), (1.0, "--")]:
        phase = sign * np.linspace(0.0 * np.pi, 1.0 * np.pi, 512)[:, np.newaxis]
        horizon = np.cos(phase) * horizon_tangent[np.newaxis, :] + np.sin(phase) * horizon_bitangent[np.newaxis, :]
        projection.plot_curve(horizon, color=0.5 * horizon_color, linestyle=style)
    # Plot the outline of a polygon
    polygon = polygon_world_space / np.linalg.norm(polygon_world_space, axis=1)[:, np.newaxis]
    outline = sample_edges(polygon, True)
    projection.plot_curve(outline, True, False, color=solid_angle_line_color)
    projection.plot_curve(outline, False, True, color=solid_angle_line_color, linestyle="--")
    # if name == "world_space":
    #     pyplot.annotate(r"$\mathbb{P}$", (-0.5, 0.7), color="w")
    # else:
    #     pyplot.annotate(r"$M^{-1}\mathbb{P}$", (-0.42, 0.75), color="k")
    # Shade the solid angle using rasterization
    x_min, x_max = (-1.0, 1.0)
    y_min, y_max = (-0.53, 1.0)
    dots_per_unit = 400.0
    width, height = (int((x_max - x_min) * dots_per_unit), int((y_max - y_min) * dots_per_unit))
    x, y = np.meshgrid(np.linspace(x_min, x_max, width), np.linspace(y_min, y_max, height))
    z = np.sqrt(1.0 - x * x - y * y)
    view_space = np.zeros(list(x.shape) + [3])
    view_space[:, :, 0] = x
    view_space[:, :, 1] = y
    view_space[:, :, 2] = z
    world_space = np.einsum("ij, ...j-> ...i", projection.matrix.T, view_space)
    cosine_space = np.einsum("ij, ...j-> ...i", world_to_cosine_space, world_space)
    cosine_space_length = np.linalg.norm(cosine_space, axis=-1)
    ltc_density = np.maximum(0.0, cosine_space[..., 2]) * np.linalg.det(world_to_cosine_space) / cosine_space_length**4
    ltc_density_rgb = cm.get_cmap("magma")(ltc_density / np.nanmax(ltc_density))[..., 0:3]
    ltc_density_rgb = np.where(np.isnan(ltc_density)[:, :, np.newaxis].repeat(3, 2), 1.0, ltc_density_rgb)
    ltc_density_rgb = np.where((cosine_space[..., 2] < 0.0)[:, :, np.newaxis].repeat(3, 2), 1.0, ltc_density_rgb)
    # Desaturate parts outside the solid angle of the polygon
    mask = np.ones_like(x, dtype=np.bool)
    for edge in range(polygon.shape[0]):
        normal = np.cross(polygon[edge], polygon[(edge + 1) % polygon.shape[0]])
        mask &= (np.einsum("...i, i -> ...", world_space, normal) >= 0.0)
    desaturation = 0.8
    ltc_density_rgb = np.where(mask[:, :, np.newaxis].repeat(3, 2), ltc_density_rgb, (1.0 - desaturation) * ltc_density_rgb + desaturation)
    pyplot.imshow(ltc_density_rgb, extent=(x_min, x_max, y_min, y_max), zorder=-1000.0, origin="lower")
    # Add a few annotations
    pyplot.annotate(r"$\int_\mathbb{P}$  " if name == "world_space" else r"$\int_{M^{-1}\mathbb{P}}$", (-0.61, 0.9), ha="right", va="center")
    pyplot.annotate(r"d$\omega\;=$" if name == "world_space" else r"d$\omega_c$", (0.56, 0.9), ha="left", va="center")
    if name == "world_space":
        pyplot.annotate(r"$M\mathbb{H}$", (-0.2, -0.45), color="k")
        pyplot.annotate(r"$p_M(\omega)$", (-0.65, 0.65), color="w")
    else:
        pyplot.annotate(r"$\mathbb{H}$", (0.2, -0.05), color="k")
        pyplot.annotate(r"$\frac{1}{\pi}\omega_{c,z}$", (-0.2, 0.8), color="k")
    # Set up axis limits
    padding = 0.02
    pyplot.xlim(-1.0 - padding, 1.0 + padding)
    pyplot.ylim(-0.53 - padding, 1.0 + padding)
    # Save as *.svg
    output_path = "../Graphics/polygon/ltc/%s" % name
    figure.savefig(output_path + ".svg", bbox_inches="tight", pad_inches=0.0)
    # Export the image as *.png without resampling
    ltc_density_rgb = np.asarray(np.clip(ltc_density_rgb, 0.0, 1.0) * 255.0, dtype=np.uint8)
    imwrite(output_path + ".png", ltc_density_rgb)
    # Now patch the *.svg file using an image without resampling
    with open(output_path + ".svg", "r") as svg_file:
        svg_code = svg_file.read()
    svg_code = re.sub(r'data:image/png;base64,[^\"]+', name + ".png", svg_code, count=1)
    with open(output_path + ".svg", "w") as svg_file:
        svg_file.write(svg_code)
    # Convert from SVG to PDF using Inkscape
    inkscape_paths = ["inkscape", "C:/Program files/Inkscape/inkscape.exe"]
    for inkscape_path in inkscape_paths:
        try:
            subprocess.call([inkscape_path, "--export-type", "pdf", "-o", output_path + ".pdf", output_path + ".svg"])
            break
        except FileNotFoundError:
            print("Failed to run Inkscape at path %s." % inkscape_path)


def plot_ltc():
    """
    Creates a plot showing how linearly transformed cosines (LTCs) are
    constructed and how they apply to linear lights.
    """
    shading_pitch = -0.6 * np.pi
    world_to_shading_space = np.asarray([
        [np.cos(shading_pitch), -np.sin(shading_pitch), 0.0],
        [np.sin(shading_pitch), np.cos(shading_pitch), 0.0],
        [0.0, 0.0, 1.0],
    ])
    height = 0.5
    polygon_world_space = np.asarray([
        [-0.5, -0.6, height],
        [0.3, -0.6, height],
        [0.3, -0.1, height],
        [-0.5, -0.1, height],
    ])
    # The code to get this transform out of ltc_table.c is:
    # if (i==34 && j == (20*resolution+26))
    # printf("[[%f, %f, %f], [%f, %f, %f], [%f, %f, %f]]\n",
    #     inverse[0][0], inverse[0][1], inverse[0][2],
    #     inverse[1][0], inverse[1][1], inverse[1][2],
    #     inverse[2][0], inverse[2][1], inverse[2][2]);
    shading_to_cosine_space = np.asarray([[0.782347, 0.000000, -0.134090], [0.000000, 1.000000, 0.000000], [0.422520, 0.000000, 0.262691]]).T
    world_to_cosine_space = shading_to_cosine_space @ world_to_shading_space
    plot_ltc_space("cosine_space", np.eye(3), polygon_world_space @ world_to_cosine_space.T)
    plot_ltc_space("world_space", world_to_cosine_space, polygon_world_space)


def plot_solid_angle_sampling():
    """
    Creates figures to visualize the inner workings of solid angle sampling for
    polygons and triangles.
    """
    # Define the polygon
    phases = np.linspace(-np.pi, np.pi, 6)[1:]
    phases = np.roll(phases, 1)
    polygon = np.vstack([np.cos(phases) + 0.4, np.sin(phases) + 0.2, 0.9 * np.ones_like(phases)]).T
    yz_rotation = 0.35 * np.pi
    rotation = np.asarray([[np.cos(yz_rotation), -np.sin(yz_rotation)], [np.sin(yz_rotation), np.cos(yz_rotation)]])
    polygon[:, 1:] = polygon[:, 1:] @ rotation.T
    # Define a subtriangle
    sampled_triangle_index = 1
    lerp_factor = 0.6
    new_vertex = (1.0 - lerp_factor) * polygon[sampled_triangle_index + 1, :] + lerp_factor * polygon[sampled_triangle_index + 2, :]

    def mask_projected_polygon(points):
        """
        :return: True for points in the solid angle of the polygon or occluded
            by it.
        """
        matrices = np.zeros(list(points.shape)[:-1] + [polygon.shape[0], 3, 3])
        matrices[..., 0, :] = points[..., np.newaxis, :]
        matrices[..., 1, :] = polygon
        matrices[..., 2, :] = np.roll(polygon, 1, axis=0)
        in_polygon = np.all(np.linalg.det(matrices) < 0.0, axis=-1)
        # Flip the points along the depth axis and test for inclusion again
        depth_axis = projection.matrix[2, :]
        flipped_points = points - 2.0 * np.einsum("i, j, ...j -> ...i", depth_axis, depth_axis, points)
        matrices[..., 0, :] = flipped_points[..., np.newaxis, :]
        in_flipped_polygon = np.all(np.linalg.det(matrices) < 0.0, axis=-1)
        # We only care about points behind the polygon
        depth = np.einsum("i, ...i -> ...", depth_axis, points)
        behind_polygon = in_flipped_polygon & (depth < 0.0)
        return in_polygon | behind_polygon

    def plot_sampled_polygon(polygon, sampled_triangle_index):
        """
        Plots the given polygon with edges separating triangles in the triangle
        fan and the area clockwise of the sample highlighted.
        """
        # Begin with the polygon outline and a fill
        outline = sample_edges(polygon, True)
        projection.plot_curve(outline, color=solid_angle_outline_color, zorder=40.0)
        screen_outline = projection.transform(outline)
        pyplot.fill(screen_outline[:, 0], screen_outline[:, 1], color=solid_angle_color)
        pyplot.tight_layout(pad=0.0, h_pad=0.0, w_pad=0.0)
        # Now plot edges separating the triangles
        for i in range(2, polygon.shape[0] - 1):
            edge = sample_edges([polygon[0, :], polygon[i, :]], False)
            projection.plot_curve(edge, color=solid_angle_line_color if i > sampled_triangle_index + 1 else sampled_area_line_color, zorder=20.0)
        # Highlight parts clockwise of a sample
        sampled_area_outline = sample_edges(np.vstack([polygon[0:sampled_triangle_index + 2, :], new_vertex[np.newaxis, :]]), True)
        screen_outline = projection.transform(sampled_area_outline)
        pyplot.fill(screen_outline[:, 0], screen_outline[:, 1], color=sampled_area_color)
        outline_edge = sample_edges(np.vstack([polygon[0, :], new_vertex[np.newaxis, :]]), True)
        projection.plot_curve(outline_edge, color=sampled_area_outline_color, zorder=30.0)

    # Width and height pertain to the first plot, xlim ylim to the second
    width, height = 2.02, 1.4
    xlim = (-0.81, 0.61)
    ylim = (-0.19, 0.7)
    # Having both, we can compute the relative size of the two plots
    aspect_0 = width / height
    aspect_1 = (xlim[1] - xlim[0]) / (ylim[1] - ylim[0])
    absolute_height = 1.0 / (aspect_0 + aspect_1)
    print(absolute_height)

    # Plot the whole triangle fan
    projection = get_euler_projection(1.85 * np.pi, 0.05 * np.pi)
    figure = pyplot.figure(figsize=get_figure_size(column_width * aspect_0 * absolute_height, column_width * absolute_height))
    pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=(-0.5 * width, 0.5 * width), ylim=(1.01 - height, 1.01))
    # Plot the outline of the unit sphere
    dense_phases = np.linspace(-np.pi, np.pi, 1001)
    circle = np.vstack([np.cos(dense_phases), np.sin(dense_phases), np.zeros_like(dense_phases)]).T
    pyplot.plot(circle[:, 0], circle[:, 1], color=horizon_color)
    # Plot the horizon
    projection.plot_curve(circle, True, False, mask_projected_polygon, color=sampled_area_line_color, linestyle="-")
    projection.plot_curve(circle, False, True, mask_projected_polygon, color=sampled_area_line_color, linestyle="--")
    projection.plot_curve(circle, True, False, lambda points: ~mask_projected_polygon(points), color=horizon_color, linestyle="-")
    projection.plot_curve(circle, False, True, lambda points: ~mask_projected_polygon(points), color=horizon_color, linestyle="--")
    # Plot the polygon
    plot_sampled_polygon(polygon, sampled_triangle_index)
    figure.savefig("../Graphics/polygon/solid_angle_sampling/triangle_fan.pdf")

    # Now plot the sampled triangle only, with a few additional annotations
    figure = pyplot.figure(figsize=get_figure_size(column_width * aspect_1 * absolute_height, column_width * absolute_height))
    pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=xlim, ylim=ylim)
    # Plot the polygon
    triangle = np.asarray([polygon[0, :], polygon[sampled_triangle_index + 1, :], polygon[sampled_triangle_index + 2, :]])
    plot_sampled_polygon(triangle, 0)
    # Annotate the vertices
    all_vertices = np.vstack([triangle, new_vertex[np.newaxis, :]])
    all_vertices /= np.linalg.norm(all_vertices, axis=-1)[:, np.newaxis]
    labels = ["$v_1$", "$v_0$", "$v_2$", r"$v_2^\prime$"]
    screen_vertices = projection.transform(all_vertices)
    offsets = [(0.01, -0.06), (0.04, -0.05), (0.05, 0.04), (0.07, 0.00)]
    for j in range(screen_vertices.shape[0]):
        pyplot.text(screen_vertices[j, 0] + offsets[j][0], screen_vertices[j, 1] + offsets[j][1], labels[j], ha="center", va="center")
    # Annotate the sizes of the solid angles
    subtriangle = np.asarray([screen_vertices[0], screen_vertices[1], screen_vertices[3]])
    area_text_position = subtriangle.mean(0)
    pyplot.text(area_text_position[0] - 0.13, area_text_position[1] + 0.13, "$A$", ha="center", va="center")
    pyplot.text(0.1, 0.47, r"$A_\Delta$", ha="center", va="center", zorder=100.0)
    figure.savefig("../Graphics/polygon/solid_angle_sampling/triangle.pdf")


def plot_sorting_networks():
    """
    Creates a figure that shows all of our sorting networks for bitonic
    sequences.
    """
    # These are the sorting networks (taken straight from the GLSL code)
    sorting_networks = [
        [(1, 2)],
        [(1, 3)],
        [(2, 4), (1, 3), (1, 2), (0, 3), (3, 4)],
        [(3, 5), (2, 4), (1, 5), (0, 4), (4, 5), (1, 3)],
        [(2, 5), (1, 6), (5, 6), (3, 4), (0, 4), (4, 6), (1, 3), (3, 5), (4, 5)],
        [(2, 6), (3, 7), (1, 5), (0, 4), (4, 6), (5, 7), (6, 7), (4, 5), (1, 3)]
    ]
    for i, network in enumerate(sorting_networks):
        # Add the shared suffix
        count = i + 3
        network += [(0, 2)]
        if count >= 4:
            network += [(2, 3)]
        network += [(0, 1)]
    print([len(network) for network in sorting_networks])
    # We plot them all into one big figure with one subplot per network
    figure = pyplot.figure(figsize=get_figure_size(column_width, column_width))
    for i, network in enumerate(sorting_networks):
        count = i + 3
        axes = pyplot.subplot(3, 2, i + 1, frameon=False, xticks=[], yticks=range(count))
        axes.invert_yaxis()
        pyplot.grid()
        x = 0.0
        # To illustrate the depth, new layers get an extra offset
        touched = count * [False]
        for comparison in network:
            if touched[comparison[0]] or touched[comparison[1]]:
                x += 1.5
                touched = count * [False]
            touched[comparison[0]] = touched[comparison[1]] = True
            pyplot.plot([x, x], comparison, "-o", color="k")
            x += 1.0
    figure.savefig("../Graphics/polygon/sorting/networks.pdf", bbox_inches="tight", pad_inches=0.03)


def plot_teaser():
    """Creates magnified insets for the teaser."""
    zoom = 8
    size = 48, 32
    pad = 24
    line_width = 16
    for source_path, name in list_experiments(r"attic_.*\.png"):
        screenshot = Screenshot(source_path)
        screenshot.add(Magnifier((1182, 904), size, (screenshot.width - pad - size[0] * zoom, pad), zoom, "C1", line_width))
        screenshot.add(Magnifier((899, 1041), size, (pad, pad), zoom, "C3", line_width))
        screenshot.add(Magnifier((552, 1283), size, (pad, screenshot.height - pad - size[1] * zoom), zoom, "C6", line_width))
        screenshot.save(path.join("../Graphics/polygon/teaser", name))


def plot_error_attic():
    """
    Defines magnified insets for screenshots showing backward error and creates
    the corresponding color map.
    """
    zoom = 8
    size = 48, 32
    pad = 24
    line_width = 16
    for source_path, name in list_experiments(r"error_attic_.*\.png"):
        screenshot = Screenshot(source_path)
        screenshot.height = 1024
        screenshot.image = screenshot.image[:screenshot.height, :]
        screenshot.add(Magnifier((1183, 945), size, (screenshot.width - pad - size[0] * zoom, pad), zoom, "C1", line_width))
        screenshot.add(Magnifier((647, 48), size, (pad, pad), zoom, "C3", line_width))
        screenshot.add(Magnifier((30, 551), size, (pad, screenshot.height - pad - size[1] * zoom), zoom, "C6", line_width))
        screenshot.save(path.join("../Graphics/polygon/error_attic", name))
    # Plot the color map
    figure = pyplot.figure(figsize=get_figure_size(0.8 * column_width, 0.04 * column_width))
    axes = pyplot.subplot()
    pyplot.colorbar(cm.ScalarMappable(norm=colors.Normalize(-7.0, -2.0), cmap="tab20b"), orientation="horizontal", format="$10^{%.0f}$", cax=axes)
    figure.savefig("../Graphics/polygon/error_attic/color_map.pdf", bbox_inches="tight", pad_inches=0.0)


def plot_small_polygon():
    """Defines magnified insets for the scene with the tiny light sources."""
    zoom = 6
    size = 64, 72
    pad = 24
    line_width = 16
    for source_path, name in list_experiments(r"bistro_(small|tiny)_polygon_.*\.png"):
        screenshot = Screenshot(source_path)
        screenshot.add(Magnifier((1545, 675), size, (screenshot.width - pad - size[0] * zoom, pad), zoom, "C1", line_width))
        screenshot.add(Magnifier((655, 309), size, (pad, pad), zoom, "C2", line_width))
        screenshot.add(Magnifier((550, 967), size, (pad, screenshot.height - pad - size[1] * zoom), zoom, "C3", line_width))
        screenshot.save(path.join("../Graphics/polygon/small_polygon", name))


def plot_mis_plane():
    """Creates magnified insets for the comparison of MIS heuristics."""
    zoom = 6
    size = 36, 36
    pad = 32
    line_width = 12
    full_list = ["mis_plane_reference.png"]
    for source_path, name in list_experiments(r"mis_plane_.*\.png"):
        screenshot = Screenshot(source_path)
        crop_y = 0
        if name not in full_list:
            crop_y = 430
            screenshot.height -= crop_y
            screenshot.image = screenshot.image[crop_y:, :]
        row_y = screenshot.height - pad - size[1] * zoom
        offset_x = (screenshot.width - size[0] * zoom - 2 * pad) // 3
        screenshot.add(Magnifier((30, 450 - crop_y), size, (pad + 0 * offset_x, row_y), zoom, "C6", line_width))
        screenshot.add(Magnifier((30, 650 - crop_y), size, (pad + 1 * offset_x, row_y), zoom, "C0", line_width))
        screenshot.add(Magnifier((400, 700 - crop_y), size, (pad + 2 * offset_x, row_y), zoom, "C3", line_width))
        screenshot.add(Magnifier((540, 650 - crop_y), size, (pad + 3 * offset_x, row_y), zoom, "C1", line_width))
        # screenshot.add(Magnifier((936, 508 - crop_y), size, (pad + 3 * offset_x, row_y), zoom, "C1", line_width))
        screenshot.save(path.join("../Graphics/polygon/mis_plane", name))


def plot_cornell_box():
    """Creates magnified insets for screenshots of the Cornell box."""
    zoom = 6
    size = 32, 32
    pad = 32
    line_width = 12
    offset_y = size[1] * zoom + pad
    for source_path, name in list_experiments(r"cornell_box_.*\.png"):
        screenshot = Screenshot(source_path)
        column_x = screenshot.width - pad - size[0] * zoom
        column_y = screenshot.height - offset_y * 3
        screenshot.add(Magnifier((365, 806), size, (pad, screenshot.height - offset_y), zoom, "C2", line_width))
        screenshot.add(Magnifier((195, 373), size, (pad, screenshot.height - 2 * offset_y), zoom, "C0", line_width))
        screenshot.add(Magnifier((522, 295), (32, 14), (column_x, column_y - pad - 14 * zoom), zoom, "C6", line_width))
        screenshot.add(Magnifier((353, 127), size, (column_x, column_y + 0 * offset_y), zoom, "C1", line_width))
        screenshot.add(Magnifier((432, 470), size, (column_x, column_y + 1 * offset_y), zoom, "C8", line_width))
        screenshot.add(Magnifier((668, 978), size, (column_x, column_y + 2 * offset_y), zoom, "C3", line_width))
        screenshot.save(path.join("../Graphics/polygon/cornell_box", name))


def plot_stratification():
    """
    Plots sample sets in several configurations to visualize stratification.
    """
    projection = get_top_projection()
    point_file = "blue_noise_de_goes.dat"
    blue_noise_points = np.fromfile(point_file, sep=" ").reshape((-1, 2))

    def plot_samples(get_sample):
        samples = np.zeros((blue_noise_points.shape[0], 3))
        for j in range(samples.shape[0]):
            samples[j, :] = get_sample(np.copy(blue_noise_points[j, :]))
        pyplot.plot(samples[:, 0], samples[:, 1], "o", color="k", fillstyle="full", mew=0.0, ms=1.7, zorder=100.0)
        # Also plot a grid
        resolution = (12, 12, 16)
        grid_0, grid_1 = np.meshgrid(np.linspace(0.0, 1.0, resolution[0] * resolution[2] + 1), np.linspace(0.0, 1.0, resolution[1] + 1), indexing="ij")
        grid_samples = np.zeros(list(grid_0.shape) + [3])
        for x in range(grid_0.shape[0]):
            for y in range(grid_0.shape[1]):
                grid_samples[x, y, :] = get_sample(np.asarray([grid_0[x, y], grid_1[x, y]]))
        for x in range(resolution[0]):
            for y in range(resolution[1]):
                if x % 2 == y % 2:
                    continue
                inner = grid_samples[x * resolution[2]:(x + 1) * resolution[2] + 1, y, 0:2]
                outer = grid_samples[x * resolution[2]:(x + 1) * resolution[2] + 1, y + 1, 0:2]
                outline = np.concatenate([inner, outer[::-1, :]], axis=0)
                pyplot.fill(outline[:, 0], outline[:, 1], color=3 * [0.75], linewidth=0.0)


    # Define two test polygons
    height = 0.5
    central_polygon = prepare_polygon(np.asarray([
        (-0.05, 0.9, height), (-0.1, -0.8, height), (0.8, -0.9, height), (0.9, 0.7, height)
    ]))
    # central_limits = [(-0.18, 1.01), (-1.01, 1.01), None]
    central_limits = [(-0.25, 1.08), (-1.01, 1.01), None]
    decentral_polygon = prepare_polygon(np.asarray([
        (-0.05, 0.9, height), (0.13, -0.6, height), (0.6, -2.3, height), (1.1, 0.7, height)
    ]))
    # decentral_limits = [(-0.06, 1.01), (-1.01, 1.01), None]
    decentral_limits = [(-0.13, 1.08), (-1.01, 1.01), None]
    # Compute appropriate sizes for the plots
    aspects = [(limits[0][1] - limits[0][0]) / (limits[1][1] - limits[1][0]) for limits in [central_limits, decentral_limits]]
    height = 1.0 / (2.0 * aspects[0] + aspects[1])
    print(height)
    central_limits[2] = (aspects[0] * height, height)
    decentral_limits[2] = (aspects[1] * height, height)

    # Solid angle sampling
    figure = pyplot.figure(figsize=get_figure_size(central_limits[2][0] * column_width, central_limits[2][1] * column_width))
    pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=central_limits[0], ylim=central_limits[1])
    plot_polygon(central_polygon, projection, ["projected_polygon_no_fill", "horizon"])
    clipped_polygon = np.vstack([central_polygon[1], central_polygon[1][0:1, :]])
    prepared_sampling = prepare_solid_angle_polygon_sampling(central_polygon[1].shape[0], clipped_polygon, np.zeros(3))
    plot_samples(lambda random_numbers: sample_solid_angle_polygon(prepared_sampling, random_numbers))
    pyplot.scatter([central_polygon[2][0, 0]], [central_polygon[2][0, 1]], marker="x", color="C3", zorder=1000.0)
    pyplot.tight_layout(pad=0.0, h_pad=0.0, w_pad=0.0)
    figure.savefig("../Graphics/polygon/stratification/solid_angle.pdf")
    # Projected solid angle sampling
    for name, prepared_polygon, limits in [("central", central_polygon, central_limits), ("decentral", decentral_polygon, decentral_limits)]:
        figure = pyplot.figure(figsize=get_figure_size(limits[2][0] * column_width, limits[2][1] * column_width))
        pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False, xlim=limits[0], ylim=limits[1])
        plot_polygon(prepared_polygon, projection, ["projected_polygon_no_fill", "horizon"])
        clipped_polygon = np.vstack([prepared_polygon[1], prepared_polygon[1][0:1, :]])
        prepared_sampling = prepare_projected_solid_angle_polygon_sampling(prepared_polygon[1].shape[0], clipped_polygon)
        plot_samples(lambda random_numbers: sample_projected_solid_angle_polygon(prepared_sampling, random_numbers))
        pyplot.scatter([0.0], [0.0], marker="x", color="C3", zorder=1000.0)
        pyplot.tight_layout(pad=0.0, h_pad=0.0, w_pad=0.0)
        figure.savefig("../Graphics/polygon/stratification/projected_solid_angle_%s.pdf" % name)


def plot_shadowed_plane_bias():
    """
    Creates magnified insets and a difference image for a scene of a shadowed
    plane that provokes bias in our biased method.
    """
    # Create a difference image
    images = [imread(source_path) for source_path, _ in list_experiments(r"shadowed_plane_.*\.hdr")]
    error_scale = 10.0
    error = np.abs(images[1] - images[0])
    error_ldr = np.asarray(linear_to_srgb(np.clip(error_scale * error, 0.0, 1.0)) * 255.0, dtype=np.uint8)
    error_path = "../Graphics/polygon/shadowed_plane_bias/error.png"
    imwrite(error_path, error_ldr)
    # Create magnified insets
    experiments = list_experiments(r"shadowed_plane_.*\.png")
    experiments.append((error_path, "error.png"))
    for source_path, name in experiments:
        screenshot = Screenshot(source_path)
        screenshot.add(Magnifier((451, 627), (64, 64), (32, 32), 6, "C1", 12))
        screenshot.save(path.join("../Graphics/polygon/shadowed_plane_bias", name))


def plot_roughness_planes_lambertian():
    """
    Creates magnified insets for the three planes of different roughness lit by
    a Lambertian emitter.
    """
    zoom = 8
    size = 48, 36
    pad = 24
    line_width = 16
    for source_path, name in list_experiments(r"roughness_planes_lambertian_.*\.png"):
        screenshot = Screenshot(source_path)
        left_column_x = pad
        right_column_x = screenshot.width - pad - size[0] * zoom
        column_y = pad
        offset_y = size[1] * zoom + 2 * pad
        screenshot.add(Magnifier((678, 249), size, (left_column_x, column_y + 0 * offset_y), zoom, "C0", line_width))
        screenshot.add(Magnifier((624, 458), size, (left_column_x, column_y + 1 * offset_y), zoom, "C1", line_width))
        screenshot.add(Magnifier((1106, 456), size, (left_column_x, column_y + 2 * offset_y), zoom, "C2", line_width))
        screenshot.add(Magnifier((1713, 219), size, (right_column_x, column_y + 0 * offset_y), zoom, "C3", line_width))
        screenshot.add(Magnifier((1767, 533), size, (right_column_x, column_y + 1 * offset_y), zoom, "C4", line_width))
        screenshot.add(Magnifier((1198, 670), size, (right_column_x, column_y + 2 * offset_y), zoom, "C5", line_width))
        screenshot.save(path.join("../Graphics/polygon/roughness_planes_lambertian", name))


def plot_ies_profile_attic():
    """
    Creates magnified insets for the attic with the high-contrast light probe.
    """
    zoom = 8
    size = 48, 36
    pad = 24
    line_width = 16
    for source_path, name in list_experiments(r"ies_profile_attic_.*\.png"):
        screenshot = Screenshot(source_path)
        column_x = screenshot.width - pad - size[0] * zoom
        column_y = pad
        offset_y = size[1] * zoom + 2 * pad
        screenshot.add(Magnifier((450, 300), size, (column_x, column_y + 0 * offset_y), zoom, "C0", line_width))
        screenshot.add(Magnifier((220, 350), size, (column_x, column_y + 1 * offset_y), zoom, "C1", line_width))
        screenshot.add(Magnifier((250, 870), size, (column_x, column_y + 2 * offset_y), zoom, "C2", line_width))
        screenshot.save(path.join("../Graphics/polygon/ies_profile_attic", name))


def plot_roughness_planes_screen():
    """
    Creates magnified insets for the three planes of different roughness lit by
    a Lambertian emitter.
    """
    zoom = 8
    size = 48, 36
    pad = 24
    line_width = 16
    for source_path, name in list_experiments(r"roughness_planes_screen_.*\.png"):
        screenshot = Screenshot(source_path)
        row_x = pad
        row_y = screenshot.height - pad - size[1] * zoom
        offset_x = (screenshot.width - 2 * pad - size[0] * zoom) // 2
        screenshot.add(Magnifier((160, 220), size, (row_x + 0 * offset_x, row_y), zoom, "C0", line_width))
        screenshot.add(Magnifier((723, 430), size, (row_x + 1 * offset_x, row_y), zoom, "C1", line_width))
        screenshot.add(Magnifier((1135, 627), size, (row_x + 2 * offset_x, row_y), zoom, "C2", line_width))
        screenshot.save(path.join("../Graphics/polygon/roughness_planes_screen", name))


def print_timing_tables(vertex_counts):
    """
    Prints a table with timings for sampling techniques applied to polygons with
    vertex counts in the given list (between 3 and 7).
    """
    # Read in the timings
    timing_dict = dict()
    for file_name in listdir("../vulkan_renderer/data/experiments"):
        match = re.match(r"timings_(de)?central_([0-9])(_128)?_(\w+)_([0-9]+\.[0-9]+)\.png", file_name)
        if match is not None:
            timing_dict[match.group(1) is not None, int(match.group(2)), match.group(3) is None, match.group(4)] = float(match.group(5))
    # The order in which techniques are listed. The boolean indicates whether
    # the timing for the decentral case is used
    techniques = [
        ("area_turk", True),
        ("rectangle_solid_angle_urena", True),
        ("solid_angle_arvo", True),
        ("solid_angle_ours", True),
        ("clipped_solid_angle_ours", True),
        ("bilinear_cosine_warp_hart", True),
        ("biquadratic_cosine_warp_hart", True),
        ("projected_solid_angle_arvo", False),
        ("projected_solid_angle_ours", False),
        ("projected_solid_angle_biased_ours", False),
        ("projected_solid_angle_arvo", True),
        ("projected_solid_angle_ours", True),
        ("projected_solid_angle_biased_ours", True),
    ]
    if 4 not in vertex_counts:
        techniques = techniques[:1] + techniques[2:]

    def format_timing(timing_in_milliseconds):
        timing = timing_in_milliseconds
        if timing < 1.0:
            return "%.2f" % timing
        elif timing < 10.0:
            return "%.2f" % timing
        elif timing < 100.0:
            return "%.1f" % timing
        else:
            return "%.0f" % timing

    # Print them line by line
    print_name = False
    for technique, case in techniques:
        if print_name:
            print(technique, end="")
        for many_samples in [False, True]:
            for vertex_count in vertex_counts:
                if technique == "rectangle_solid_angle_urena" and vertex_count != 4:
                    print("-\t", end="")
                else:
                    print("%s\t" % format_timing(timing_dict[case, vertex_count, many_samples, technique] - timing_dict[True, vertex_count, many_samples, "baseline"]), end="")
        print()
    if print_name:
        print("baseline", end="")
    for many_samples in [False, True]:
        for vertex_count in vertex_counts:
            print("%s\t" % format_timing(timing_dict[True, vertex_count, many_samples, "baseline"]), end="")
    print()


if __name__ == "__main__":
    # plot_geometric_constructions()
    # plot_case_distinction()
    # plot_iteration()
    # plot_initizialization()
    # plot_arvo_failure()
    # plot_ltc()
    # plot_solid_angle_sampling()
    # plot_sorting_networks()
    # plot_teaser()
    # plot_error_attic()
    # plot_small_polygon()
    # plot_mis_plane()
    # plot_cornell_box()
    # plot_stratification()
    # plot_shadowed_plane_bias()
    # plot_roughness_planes_lambertian()
    # plot_ies_profile_attic()
    # plot_roughness_planes_screen()
    # print_experiment_errors("attic_.*\.hdr$", "attic_reference.hdr", "rmse")
    # print_experiment_errors("mis_plane_.*\.hdr$", "mis_plane_reference.hdr", "rmse")
    # print_experiment_errors("mis_plane_.*\.hdr$", "mis_plane_reference.hdr", "clamped_rmse")
    # print_experiment_errors("cornell_box_.*\.hdr$", "cornell_box_reference.hdr", "rmse")
    # print_experiment_errors("cornell_box_.*_tilted.*\.hdr$", "cornell_box_reference_tilted.hdr", "rmse")
    # print_timing_tables([3, 4, 5])
    # print_timing_tables([6, 7])
    print("Done.")
