#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
import matplotlib
from matplotlib import pyplot, path, markers

# Tolerance where a value still counts as zero for the sake of clipping
clipping_epsilon = 1.0e-6


class OrthogonalProjection:
    """
    Encapsulates a 3x3 orthogonal matrix and provides functionality to use it
    as a world to projection space matrix for matplotlib plots.
    """

    def __init__(self, orthogonal_matrix_3x3):
        """Sets the projection to the given orthogonal 3x3 matrix."""
        self.matrix = np.asarray(orthogonal_matrix_3x3).copy()

    def plot_curve(self, curve, plot_front=True, plot_back=True, mask_functor=None, **keyword_arguments):
        """
        Given one or more curves in 3D as array broadcastable to shape
        (curve_count, sample_count, 3), this function transforms it to viewport
        coordinates and plots it. Optionally front or back parts are masked out
        (the distinguishing plane has constant depth and goes through the
        origin). You can also pass a Functor taking a (...,3) world space array
        that decides what gets masked (False to exclude). Keyword arguments
        forward to pyplot.plot.
        """
        # Ensure that we have a 3D array
        if curve.ndim == 2:
            curve = curve[np.newaxis, :, :]
        view_curve = self.transform(curve)
        mask = np.zeros(curve[..., 0].shape, dtype=np.bool)
        if plot_front:
            mask |= view_curve[..., 2] >= -clipping_epsilon
        if plot_back:
            mask |= view_curve[..., 2] <= -clipping_epsilon
        if mask_functor is not None:
            mask &= mask_functor(np.asarray(curve))
        if not np.any(mask):
            return
        # Now flatten the array with a masked out vertex separating different
        # curves
        padded_view_curve = np.concatenate([view_curve, np.zeros((curve.shape[0], 1, 3))], axis=1)
        padded_mask = np.concatenate([mask, np.zeros((curve.shape[0], 1), dtype=bool)], axis=1)
        view_curve = padded_view_curve.reshape((-1, 3))
        mask = padded_mask.reshape((-1,))
        # Use NaNs to mask out parts
        view_curve[~mask, :] = np.NAN
        # Plot
        pyplot.plot(view_curve[:, 0], view_curve[:, 1], **keyword_arguments)

    def plot_point_set(self, points, marker_size=1.0, **keyword_arguments):
        """
        Given an array of shape (...,3) in 3D space, this function plots it
        using this projection. The markers are circles of specified size.
        keyword_arguments forward to pyplot.scatter().
        """
        if points.size == 0:
            return
        transformed_points = self.transform(points)
        marker_phase = np.linspace(-np.pi, np.pi, 16)
        marker_vertices = np.vstack([np.cos(marker_phase), np.sin(marker_phase)]).T
        marker_path = matplotlib.path.Path(marker_vertices, closed=True)
        default_keyword_arguments = dict(marker=markers.MarkerStyle(marker_path, "top"), s=marker_size, linewidths=0.0)
        pyplot.scatter(transformed_points[..., 0].flat, transformed_points[..., 1].flat,
                       **merge_dict(default_keyword_arguments, keyword_arguments))

    def transform(self, point):
        """
        Given an array of shape (...,3), this function applies the projection
        transform and returns an array of same shape where [...,0] and [...,1]
        hold the coordinates for a plot.
        """
        return np.einsum("ij, ...j -> ...i", self.matrix, point)

    def inverse_transform(self, point):
        """Like transform() except that it applies the inverse transform."""
        return np.einsum("ji, ...j -> ...i", self.matrix, point)


def get_top_projection():
    """Returns a projection that shows the situation seen from the top."""
    return OrthogonalProjection([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])


def get_euler_projection(yaw, pitch):
    """
    Constructs a projection using Euler rotation angles.
    :return: An OrthogonalProjection.
    """
    perspective_rotation = np.asarray([
        [1.0, 0.0, 0.0],
        [0.0, np.cos(pitch), -np.sin(pitch)],
        [0.0, np.sin(pitch), np.cos(pitch)]])
    perspective_rotation = perspective_rotation @ np.asarray([
        [np.cos(yaw), -np.sin(yaw), 0.0],
        [0.0, 0.0, 1.0],
        [-np.sin(yaw), -np.cos(yaw), 0.0]])
    return OrthogonalProjection(perspective_rotation)


def merge_dict(*argument_list):
    """
    A little helper to merge dictionaries. If a key is in multiple
    dictionaries, the returned dictionary uses the value of the last one in the
    list.
    """
    result = dict()
    for argument in argument_list:
        result.update(argument)
    return result


def plot_hemisphere(orthogonal_projection, longitude_count=8, latitude_count=4, mask_functor=None, sample_count=257,
                    **keyword_arguments):
    """
    Plots the unit hemisphere as collection of lines. mask_functor forwards to
    OrthogonalProjection.plot_curve(). keyword_arguments forward to every
    invocation of pyplot.plot.
    """
    # Plot lines with constant longitude
    arc_phase = np.linspace(0.0, np.pi, sample_count)
    arc_x = np.cos(arc_phase)
    arc_y = np.sin(arc_phase)
    rotations = np.linspace(0.0, np.pi, longitude_count + 1)[0:longitude_count]
    longitudes = np.dstack([
        np.cos(rotations)[:, np.newaxis] * arc_x[np.newaxis, :],
        np.sin(rotations)[:, np.newaxis] * arc_x[np.newaxis, :],
        np.ones_like(rotations)[:, np.newaxis] * arc_y[np.newaxis, :]])
    orthogonal_projection.plot_curve(longitudes, False, True, mask_functor,
                                     **merge_dict(dict(color=3 * [0.9], linestyle="--", zorder=-30.0),
                                                  keyword_arguments))
    orthogonal_projection.plot_curve(longitudes, True, False, mask_functor,
                                     **merge_dict(dict(color=3 * [0.8], zorder=-29.0), keyword_arguments))
    # Plot lines with constant latitude
    phase = np.linspace(-np.pi, np.pi, sample_count)
    circle_x = np.cos(phase)
    circle_y = np.sin(phase)
    elevations = np.linspace(0.0, 1.0, latitude_count + 1)[0:latitude_count]
    elevations = np.sqrt(elevations)
    radii = np.sqrt(1.0 - elevations * elevations)
    latitudes = np.dstack([
        radii[:, np.newaxis] * circle_x[np.newaxis, :],
        radii[:, np.newaxis] * circle_y[np.newaxis, :],
        elevations[:, np.newaxis] * np.ones_like(circle_x)[np.newaxis, :]
    ])
    for brightness_factor, range_begin, range_end in [(0.85, 0, 1), (1.0, 1, elevations.size)]:
        front_color = brightness_factor * np.asarray(keyword_arguments.get("color", 3 * [0.9]))
        orthogonal_projection.plot_curve(latitudes[range_begin:range_end, :, :], False, True, mask_functor,
                                         **merge_dict(dict(linestyle="--", zorder=-30.0), keyword_arguments,
                                                      dict(color=front_color)))
        orthogonal_projection.plot_curve(latitudes[range_begin:range_end, :, :], True, False, mask_functor,
                                         **merge_dict(dict(zorder=-29.0), dict(color=0.9 * front_color),
                                                      keyword_arguments))


def plot_annotated_length(vertical, interval_begin, interval_end, offset, label, begin_label="", end_label="",
                          line_keyword_arguments=dict(), label_offset=0.07, **keyword_arguments):
    """
    Plots an annotated line segment. It is either horizontal or vertical
    dependent on the given boolean, covers the given interval and the other
    coordinate is fixed. It is labeled with a string. keyword_arguments forward
    to pyplot.text(), LineKeywordArguments to pyplot.plot().
    """
    marker_radius = 0.05
    x = [[interval_begin, interval_end], 2 * [interval_begin], 2 * [interval_end]]
    y = [[offset, offset]] + 2 * [[offset - marker_radius, offset + marker_radius]]
    x = np.asarray(x).T
    y = np.asarray(y).T
    label_x = 0.5 * (interval_begin + interval_end)
    label_y = offset + label_offset
    begin_label_x, begin_label_y = interval_begin, label_y
    end_label_x, end_label_y = interval_end, label_y
    if vertical:
        x, y = y, x
        label_x, label_y = label_y, label_x
        begin_label_x, begin_label_y = begin_label_y, begin_label_x
        end_label_x, end_label_y = end_label_y, end_label_x
    pyplot.plot(x, y, **merge_dict(dict(color="k"), line_keyword_arguments))
    default_keyword_arguments = {"horizontalalignment": "left" if vertical else "center",
                                 "verticalalignment": "center" if vertical else "bottom"}
    pyplot.text(label_x, label_y, label, **merge_dict(default_keyword_arguments, keyword_arguments))
    pyplot.text(begin_label_x, begin_label_y, begin_label, **merge_dict(default_keyword_arguments, keyword_arguments))
    pyplot.text(end_label_x, end_label_y, end_label, **merge_dict(default_keyword_arguments, keyword_arguments))


def plot_annotated_point(x, y, label, horizontal_alignment="left", vertical_alignment="top", label_offset=(0.0, 0.0)):
    """Plots a point and annotates it with the given label."""
    pyplot.scatter([x], [y], zorder=10.0, color="k", s=1.0)
    if horizontal_alignment == "center" and vertical_alignment == "top":
        label_offset[1] -= 0.03
    pyplot.text(x + label_offset[0], y + label_offset[1], label, horizontalalignment=horizontal_alignment,
                verticalalignment=vertical_alignment)


def plot_axis(projection, origin, axis_index, other_side=False, label_offset=0.15, zorder=10000.0):
    """
    Plots the coordinate axis with the given index using the given world space
    origin. The label can be flipped to the other side if it looks bad. The
    plot may be skipped if the axis projects to depth.
    """
    origin = np.asarray(origin)
    transformed_origin = projection.transform(origin)
    name = "xyz"[axis_index]
    transformed_vector = projection.transform(np.eye(3)[:, axis_index])
    if np.linalg.norm(transformed_vector[0:2]) > 1.0e-3:
        pyplot.arrow(transformed_origin[0], transformed_origin[1], transformed_vector[0], transformed_vector[1],
                     head_width=0.08, facecolor="w", zorder=zorder)
        if label_offset is not None:
            offset_scale = (-1.0 if other_side else 1.0) * label_offset
            text_position = transformed_origin[0:2] + 0.5 * transformed_vector[0:2] + offset_scale * np.asarray([-transformed_vector[1], transformed_vector[0]])
            pyplot.text(text_position[0], text_position[1], r"$\mathbf{%s}$" % name, ha="center", va="center", zorder=zorder)


def plot_right_triangle(projection, points, hypothenuse_label, adjacent_label, opposite_label, angle_size=0.0,
                        angle_label="", right_angle_size=0.1, label_offset=(0.07, 0.07, 0.07)):
    """
    Plots a triangle in 3D space given by the (3,3) array points (last
    dimension corresponds to the three coordinates). Each edge is labeled. If
    the points are A, B, C, the right angle is at B and BC is the opposite.
    """
    transformed_points = projection.transform(points)
    x = transformed_points[:, 0]
    y = transformed_points[:, 1]
    pyplot.plot([x[0], x[1], x[2], x[0]], [y[0], y[1], y[2], y[0]], color="k")
    # Compute edge centers
    edge_x = np.roll(x, 1) - x
    edge_y = np.roll(y, 1) - y
    label_x = 0.5 * (x + np.roll(x, 1)) - edge_y * np.asarray(label_offset) / np.sqrt(edge_x ** 2 + edge_y ** 2)
    label_y = 0.5 * (y + np.roll(y, 1)) + edge_x * np.asarray(label_offset) / np.sqrt(edge_x ** 2 + edge_y ** 2)
    for i, label in enumerate([hypothenuse_label, adjacent_label, opposite_label]):
        pyplot.text(label_x[i], label_y[i], label, horizontalalignment="center", verticalalignment="center")
    # Annotate the relevant angles
    angle_list = [(1, 0, 1, 2, right_angle_size, None)]
    if angle_size > 0.0:
        angle_list += [(2, 1, 2, 0, angle_size, angle_label)]
    for point, edge_0, edge_1, edge_2, size, label in angle_list:
        orientation = np.linalg.det([[edge_x[edge_0], edge_y[edge_0]], [edge_x[edge_1], edge_y[edge_1]]])
        start_angle = -np.pi + np.angle(-edge_x[edge_1] - 1.0j * edge_y[edge_1])
        end_angle = np.angle(-edge_x[edge_2] - 1.0j * edge_y[edge_2])
        while start_angle - end_angle > (0.0 if orientation < 0.0 else np.pi):
            end_angle += np.pi
        while start_angle - end_angle < (-np.pi if orientation < 0.0 else 0.0):
            end_angle -= np.pi
        phase = np.linspace(start_angle, end_angle, 32)
        right_angle = size * np.exp(1.0j * phase)
        pyplot.plot(x[point] + right_angle.real, y[point] + right_angle.imag, color="k")
        marker_phase = np.linspace(-np.pi, np.pi, 32)
        marker_path = matplotlib.path.Path(np.vstack([np.cos(marker_phase), np.sin(marker_phase)]).T, closed=True)
        right_angle_dot_x = x[point] + 0.6 * size * np.cos(phase[15])
        right_angle_dot_y = y[point] + 0.6 * size * np.sin(phase[15])
        label_x = x[point] + (size - 0.07) * np.cos(phase[15])
        label_y = y[point] + (size - 0.07) * np.sin(phase[15])
        if label is None:
            pyplot.scatter([right_angle_dot_x], [right_angle_dot_y], zorder=100.0,
                           marker=markers.MarkerStyle(marker_path, "top"), s=6.0, linewidths=0.0, color="k")
        else:
            pyplot.text(label_x, label_y, label, horizontalalignment="center", verticalalignment="center")
