#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from glsl_math import *
MIN_POLYGONAL_LIGHT_VERTEX_COUNT = 3
MAX_POLYGONAL_LIGHT_VERTEX_COUNT = 4
MAX_POLYGON_VERTEX_COUNT = 5



class solid_angle_polygon_t:
    def __init__(self):
        self.vertex_count = 0
        self.vertex_dirs = [zeros((3,)) for _i in range(MAX_POLYGON_VERTEX_COUNT)]
        self.triangle_parameters = [zeros((3,)) for _i in range(MAX_POLYGON_VERTEX_COUNT - 2)]
        self.fan_solid_angles = [0.0 for _i in range(MAX_POLYGON_VERTEX_COUNT - 2)]
        self.solid_angle = 0.0


def fast_atan(y):
    rx = where((abs(y) > 1.0) ,(1.0 / abs(y)) ,abs(y))
    ry = rx * rx
    rz = fma(ry, 0.02083509974181652, -0.08513300120830536)
    rz = fma(ry, rz, 0.18014100193977356)
    rz = fma(ry, rz, -0.3302994966506958)
    ry = fma(ry, rz, 0.9998660087585449)
    rz = fma(-2.0 * ry, rx, 0.5 * pi)
    rz = where((abs(y) > 1.0) ,rz ,0.0)
    rx = fma(rx, ry, rz)
    return  where((y < 0.0) ,-rx ,rx)


def positive_atan(tangent):
    offset = where((tangent < 0.0) ,pi ,0.0)
    return arctan(tangent) + offset


def prepare_solid_angle_polygon_sampling(vertex_count, vertices, shading_position):
    polygon = solid_angle_polygon_t()
    polygon.vertex_count = vertex_count
    for i in range(0, MAX_POLYGON_VERTEX_COUNT):
        polygon.vertex_dirs[i] = normalize(vertices[i] - shading_position)
    householder_sign = where((polygon.vertex_dirs[0][0] > 0.0) ,-1.0 ,1.0)
    householder_yz = polygon.vertex_dirs[0][1:3] * (1.0 / (abs(polygon.vertex_dirs[0][0]) + 1.0))
    polygon.solid_angle = 0.0
    previous_dot_1_2 = dot(polygon.vertex_dirs[0], polygon.vertex_dirs[1])
    for i in range(0, MAX_POLYGON_VERTEX_COUNT - 2):
        if (i >= 1 and i + 2 >= vertex_count):
            break
        vertices = [
            polygon.vertex_dirs[i + 1],
            polygon.vertex_dirs[0],
            polygon.vertex_dirs[i + 2]]
        dot_0_1 = previous_dot_1_2
        dot_0_2 = dot(vertices[0], vertices[2])
        dot_1_2 = dot(vertices[1], vertices[2])
        previous_dot_1_2 = dot_1_2
        dot_householder_0 = fma(-householder_sign, vertices[0][0], dot_0_1)
        dot_householder_2 = fma(-householder_sign, vertices[2][0], dot_1_2)
        bottom_right_minor = mat2(
            fma(vec2(-dot_householder_0), householder_yz, vertices[0][1:3]),
            fma(vec2(-dot_householder_2), householder_yz, vertices[2][1:3]))
        simplex_volume = abs(linalg.det(bottom_right_minor))
        dot_0_2_plus_1_2 = dot_0_2 + dot_1_2
        one_plus_dot_0_1 = 1.0 + dot_0_1
        tangent = simplex_volume / (one_plus_dot_0_1 + dot_0_2_plus_1_2)
        triangle_solid_angle = 2.0 * positive_atan(tangent)
        polygon.solid_angle += triangle_solid_angle
        polygon.fan_solid_angles[i] = polygon.solid_angle
        polygon.triangle_parameters[i] = vec3(simplex_volume, dot_0_2_plus_1_2, one_plus_dot_0_1)
    return polygon


def mix_fma(x, y, a):
    return fma(a, y, fma(-a, x, x))


def sample_solid_angle_polygon(polygon, random_numbers):
    target_solid_angle = polygon.solid_angle * random_numbers[0]
    subtriangle_solid_angle = target_solid_angle
    parameters = polygon.triangle_parameters[0]
    vertices = [
        polygon.vertex_dirs[1], polygon.vertex_dirs[0], polygon.vertex_dirs[2]
    ]
    for i in range(0, MAX_POLYGON_VERTEX_COUNT - 3):
        if (i + 3 >= polygon.vertex_count or polygon.fan_solid_angles[i] >= target_solid_angle):
            break
        subtriangle_solid_angle = target_solid_angle - polygon.fan_solid_angles[i]
        vertices[0] = polygon.vertex_dirs[i + 2]
        vertices[2] = polygon.vertex_dirs[i + 3]
        parameters = polygon.triangle_parameters[i + 1]
    cos_sin = vec2(cos(0.5 * subtriangle_solid_angle), sin(0.5 * subtriangle_solid_angle))
    offset = vertices[0] * (parameters[0] * cos_sin[0] - parameters[1] * cos_sin[1]) + vertices[2] * (parameters[2] * cos_sin[1])
    new_vertex_2 = fma(2.0 * vec3(dot(vertices[0], offset) / dot(offset, offset)), offset, -vertices[0])
    s2 = dot(vertices[1], new_vertex_2)
    s = mix_fma(1.0, s2, random_numbers[1])
    denominator = fma(-s2, s2, 1.0)
    t_normed = sqrt(fma(-s, s, 1.0) / denominator)
    t_normed = where((denominator > 0.0) ,t_normed ,random_numbers[1])
    return fma(-t_normed, s2, s) * vertices[1] + t_normed * new_vertex_2


class projected_solid_angle_polygon_t:
    def __init__(self):
        self.vertex_count = 0
        self.vertices = [zeros((2,)) for _i in range(MAX_POLYGON_VERTEX_COUNT)]
        self.ellipses = [zeros((2,)) for _i in range(MAX_POLYGON_VERTEX_COUNT)]
        self.inner_ellipse_0 = zeros((2,))
        self.sector_projected_solid_angles = [0.0 for _i in range(MAX_POLYGON_VERTEX_COUNT)]
        self.projected_solid_angle = 0.0


def kahan(a, b, c, d):
    cd = c * d
    error = fma(c, d, -cd)
    result = fma(a, b, -cd)
    return result - error


def cross_stable(lhs, rhs):
    return vec3(
        kahan(lhs[1], rhs[2], lhs[2], rhs[1]),
        kahan(lhs[2], rhs[0], lhs[0], rhs[2]),
        kahan(lhs[0], rhs[1], lhs[1], rhs[0])
    )


def rotate_90(input_vector):
    return vec2(-input_vector[1], input_vector[0])


def is_inner_ellipse(ellipse):
    return (floatBitsToUint(ellipse[0]) & 0x80000000) != 0


def is_central_case(polygon):
    return polygon.inner_ellipse_0[0] > 0.0


def ellipse_from_edge(vertex_0, vertex_1):
    normal = cross_stable(vertex_0, vertex_1)
    scaling = 1.0 / normal[2]
    scaling = where(is_inner_ellipse(normal[0:2]) ,-scaling ,scaling)
    ellipse = normal[0:2] * scaling
    ellipse[0] = where((normal[2] != 0.0) ,ellipse[0] ,inf)
    return ellipse


def ellipse_transform(ellipse, point):
    return fma(vec2(dot(ellipse, point)), ellipse, point)


def get_ellipse_det(ellipse):
    return fma(ellipse[0], ellipse[0], fma(ellipse[1], ellipse[1], 1.0))

def get_ellipse_rsqrt_det(ellipse):
    return inversesqrt(get_ellipse_det(ellipse))

def get_ellipse_direction_factor_rsq(ellipse, dir):
    ellipse_dot_dir = dot(ellipse, dir)
    dir_dot_dir = dot(dir, dir)
    return fma(ellipse_dot_dir, ellipse_dot_dir, dir_dot_dir)

def get_ellipse_direction_factor(ellipse, dir):
    return inversesqrt(get_ellipse_direction_factor_rsq(ellipse, dir))

def get_ellipse_normalized_direction_factor(ellipse, normalized_dir):
    ellipse_dot_dir = dot(ellipse, normalized_dir)
    return inversesqrt(fma(ellipse_dot_dir, ellipse_dot_dir, 1.0))


def get_area_between_ellipses_in_sector_from_tangents(inner_rsqrt_det, inner_tangent, outer_rsqrt_det, outer_tangent):
    inner_area = inner_rsqrt_det * positive_atan(inner_tangent)
    result = fma(outer_rsqrt_det, positive_atan(outer_tangent), -inner_area)
    return  where((result > 0.0) ,(0.5 * result) ,0.0)


def get_area_between_ellipses_in_sector(inner_ellipse, inner_rsqrt_det, outer_ellipse, outer_rsqrt_det, dir_0, dir_1):
    det_dirs = maximum(+0.0, dot(dir_1, rotate_90(dir_0)))
    inner_dot = inner_rsqrt_det * dot(dir_0, ellipse_transform(inner_ellipse, dir_1))
    outer_dot = outer_rsqrt_det * dot(dir_0, ellipse_transform(outer_ellipse, dir_1))
    return get_area_between_ellipses_in_sector_from_tangents(
        inner_rsqrt_det, det_dirs / inner_dot,
        outer_rsqrt_det, det_dirs / outer_dot)


def get_ellipse_area_in_sector(ellipse, dir_0, dir_1):
    ellipse_rsqrt_det = get_ellipse_rsqrt_det(ellipse)
    det_dirs = maximum(+0.0, dot(dir_1, rotate_90(dir_0)))
    ellipse_dot = ellipse_rsqrt_det * dot(dir_0, ellipse_transform(ellipse, dir_1))
    area = 0.5 * ellipse_rsqrt_det * positive_atan(det_dirs / ellipse_dot)
    return  where((ellipse_rsqrt_det > 0.0) ,area ,0.0)


def compare_and_swap(polygon, lhs, rhs):
    lhs_copy = polygon.vertices[lhs]
    normal_z = kahan(lhs_copy[0], -polygon.vertices[rhs][1], lhs_copy[1], -polygon.vertices[rhs][0])
    swap = where((normal_z == 0.0) ,isinf(polygon.ellipses[rhs][0]) ,(normal_z > 0.0))
    polygon.vertices[lhs] = where(swap ,polygon.vertices[rhs] ,lhs_copy)
    polygon.vertices[rhs] = where(swap ,lhs_copy ,polygon.vertices[rhs])
    lhs_copy = polygon.ellipses[lhs]
    polygon.ellipses[lhs] = where(swap ,polygon.ellipses[rhs] ,lhs_copy)
    polygon.ellipses[rhs] = where(swap ,lhs_copy ,polygon.ellipses[rhs])


def sort_convex_polygon_vertices(polygon):
    if (polygon.vertex_count == 3):
        compare_and_swap(polygon, 1, 2)
#if MAX_POLYGON_VERTEX_COUNT >= 4
    elif (polygon.vertex_count == 4):
        compare_and_swap(polygon, 1, 3)
#endif
#if MAX_POLYGON_VERTEX_COUNT >= 5
    elif (polygon.vertex_count == 5):
        compare_and_swap(polygon, 2, 4)
        compare_and_swap(polygon, 1, 3)
        compare_and_swap(polygon, 1, 2)
        compare_and_swap(polygon, 0, 3)
        compare_and_swap(polygon, 3, 4)
#endif
#if MAX_POLYGON_VERTEX_COUNT >= 6
    elif (polygon.vertex_count == 6):
        compare_and_swap(polygon, 3, 5)
        compare_and_swap(polygon, 2, 4)
        compare_and_swap(polygon, 1, 5)
        compare_and_swap(polygon, 0, 4)
        compare_and_swap(polygon, 4, 5)
        compare_and_swap(polygon, 1, 3)
#endif
#if MAX_POLYGON_VERTEX_COUNT >= 7
    elif (polygon.vertex_count == 7):
        compare_and_swap(polygon, 2, 5)
        compare_and_swap(polygon, 1, 6)
        compare_and_swap(polygon, 5, 6)
        compare_and_swap(polygon, 3, 4)
        compare_and_swap(polygon, 0, 4)
        compare_and_swap(polygon, 4, 6)
        compare_and_swap(polygon, 1, 3)
        compare_and_swap(polygon, 3, 5)
        compare_and_swap(polygon, 4, 5)
#endif
#if MAX_POLYGON_VERTEX_COUNT >= 8
    elif (polygon.vertex_count == 8):
        compare_and_swap(polygon, 2, 6)
        compare_and_swap(polygon, 3, 7)
        compare_and_swap(polygon, 1, 5)
        compare_and_swap(polygon, 0, 4)
        compare_and_swap(polygon, 4, 6)
        compare_and_swap(polygon, 5, 7)
        compare_and_swap(polygon, 6, 7)
        compare_and_swap(polygon, 4, 5)
        compare_and_swap(polygon, 1, 3)
#endif
    compare_and_swap(polygon, 0, 2)
#if MAX_POLYGON_VERTEX_COUNT >= 4
    if (polygon.vertex_count >= 4):
        compare_and_swap(polygon, 2, 3)
#endif
    compare_and_swap(polygon, 0, 1)


def prepare_projected_solid_angle_polygon_sampling(vertex_count, vertices):
    polygon = projected_solid_angle_polygon_t()
    polygon.vertex_count = vertex_count
    polygon.inner_ellipse_0 = vec2(1.0, 0.0)
    polygon.vertices[0] = vertices[0][0:2]
    polygon.ellipses[0] = ellipse_from_edge(vertices[0], vertices[1])
    previous_ellipse = polygon.ellipses[0]
    for i in range(1, MAX_POLYGON_VERTEX_COUNT):
        polygon.vertices[i] = vertices[i][0:2]
        if (i > 2 and i == polygon.vertex_count):
            break
        ellipse = ellipse_from_edge(vertices[i], vertices[(i + 1) % MAX_POLYGON_VERTEX_COUNT])
        ellipse_inner = is_inner_ellipse(ellipse)
        polygon.ellipses[i] = where(ellipse_inner ,previous_ellipse ,ellipse)
        polygon.inner_ellipse_0 = where((is_inner_ellipse(previous_ellipse) and not ellipse_inner) ,previous_ellipse ,polygon.inner_ellipse_0)
        previous_ellipse = ellipse
    ellipse = polygon.ellipses[0]
    ellipse_inner = is_inner_ellipse(ellipse)
    polygon.ellipses[0] = where(ellipse_inner ,previous_ellipse ,ellipse)
    polygon.inner_ellipse_0 = where((is_inner_ellipse(previous_ellipse) and not ellipse_inner) ,previous_ellipse ,polygon.inner_ellipse_0)
    polygon.projected_solid_angle = 0.0
    if (is_central_case(polygon)):
        for i in range(0, MAX_POLYGON_VERTEX_COUNT):
            if (i > 2 and i == polygon.vertex_count):
                break
            polygon.sector_projected_solid_angles[i] = get_ellipse_area_in_sector(polygon.ellipses[i], polygon.vertices[i], polygon.vertices[(i + 1) % MAX_POLYGON_VERTEX_COUNT])
            polygon.projected_solid_angle += polygon.sector_projected_solid_angles[i]
    else:
        sort_convex_polygon_vertices(polygon)
        inner_ellipse = polygon.inner_ellipse_0
        inner_rsqrt_det = get_ellipse_rsqrt_det(inner_ellipse)
        outer_ellipse = zeros((2,))
        for i in range(0, MAX_POLYGON_VERTEX_COUNT - 1):
            if (i > 1 and i + 1 == polygon.vertex_count):
                break
            vertex_ellipse = polygon.ellipses[i]
            vertex_inner = is_inner_ellipse(vertex_ellipse)
            vertex_rsqrt_det = get_ellipse_rsqrt_det(vertex_ellipse)
            if (i == 0):
                outer_ellipse = vertex_ellipse
                outer_rsqrt_det = vertex_rsqrt_det
            else:
                inner_ellipse = where(vertex_inner ,vertex_ellipse ,inner_ellipse)
                inner_rsqrt_det = where(vertex_inner ,vertex_rsqrt_det ,inner_rsqrt_det)
                outer_ellipse = where(vertex_inner ,outer_ellipse ,vertex_ellipse)
                outer_rsqrt_det = where(vertex_inner ,outer_rsqrt_det ,vertex_rsqrt_det)
            polygon.sector_projected_solid_angles[i] = get_area_between_ellipses_in_sector(
                inner_ellipse, inner_rsqrt_det, outer_ellipse, outer_rsqrt_det, polygon.vertices[i], polygon.vertices[i + 1])
            polygon.projected_solid_angle += polygon.sector_projected_solid_angles[i]
    return polygon


def normalize_approx_and_flip(rhs, semi_circle):
    scaling = abs(rhs[0]) + abs(rhs[1])
    scaling = uintBitsToFloat(floatBitsToUint(scaling) ^ 0x7F800000)
    scaling = where((dot(rhs, semi_circle) >= 0.0) ,scaling ,-scaling)
    return scaling * rhs


def solve_homogeneous_quadratic(quadratic):
    coeff_xy = 0.5 * (quadratic[1][0] + quadratic[0][1])
    sqrt_discriminant = sqrt(maximum(0.0, coeff_xy * coeff_xy - quadratic[0][0] * quadratic[1][1]))
    scaled_root = abs(coeff_xy) + sqrt_discriminant
    return  where((coeff_xy >= 0.0) ,vec2(scaled_root, -quadratic[0][0]) ,vec2(quadratic[1][1], scaled_root))


def sample_sector_between_ellipses(random_numbers, target_area, inner_ellipse, outer_ellipse, dir_0, dir_1, iteration_count):
    quad_dirs = [zeros((2,)) for _i in range(3)]
    quad_dirs[0] = normalize(dir_0)
    quad_dirs[2] = normalize(dir_1)
    quad_dirs[1] = quad_dirs[0] + quad_dirs[2]
    normalization_factor = asarray([
        [
            get_ellipse_normalized_direction_factor(inner_ellipse, quad_dirs[0]),
            get_ellipse_direction_factor(inner_ellipse, quad_dirs[1]),
            get_ellipse_normalized_direction_factor(inner_ellipse, quad_dirs[2])
        ],
        [
            get_ellipse_normalized_direction_factor(outer_ellipse, quad_dirs[0]),
            get_ellipse_direction_factor(outer_ellipse, quad_dirs[1]),
            get_ellipse_normalized_direction_factor(outer_ellipse, quad_dirs[2])
        ]
    ]).T
    sector_areas = [
        normalization_factor[0][1] * normalization_factor[1][1] - normalization_factor[0][0] * normalization_factor[1][0],
        normalization_factor[1][1] * normalization_factor[2][1] - normalization_factor[1][0] * normalization_factor[2][0]
    ]
    target_quad_area = mix_fma(-sector_areas[0], sector_areas[1], random_numbers[0])
    quad_dirs[2] = where((target_quad_area <= 0.0) ,quad_dirs[0] ,quad_dirs[2])
    normalization_factor[2][0] = where((target_quad_area <= 0.0) ,normalization_factor[0][0] ,normalization_factor[2][0])
    normalization_factor[2][1] = where((target_quad_area <= 0.0) ,normalization_factor[0][1] ,normalization_factor[2][1])
    target_quad_area += where((target_quad_area <= 0.0) ,sector_areas[0] ,-sector_areas[1])
    target_quad_area *= abs(linalg.det(mat2(quad_dirs[1], quad_dirs[2])))
    quad_normals = [
        quad_dirs[1] * normalization_factor[1][0] + quad_dirs[2] * normalization_factor[2][0],
        quad_dirs[1] * normalization_factor[1][1] + quad_dirs[2] * normalization_factor[2][1]
    ]
    quad_normals[0] = ellipse_transform(inner_ellipse, quad_normals[0])
    quad_normals[1] = ellipse_transform(outer_ellipse, quad_normals[1])
    quad_offsets = [
        dot(quad_normals[0], quad_dirs[1]) * normalization_factor[1][0],
        dot(quad_normals[1], quad_dirs[1]) * normalization_factor[1][1]
    ]
    quadratic = outer((quad_offsets[1] * normalization_factor[2][1]) * rotate_90(quad_dirs[2]), quad_normals[0])
    quadratic -= outer((quad_offsets[0] * normalization_factor[2][0]) * rotate_90(quad_dirs[2]) + target_quad_area * quad_normals[0], quad_normals[1])
    current_dir = solve_homogeneous_quadratic(quadratic)

    acceptable_error = 1.0e-5
    iteration_count = where((abs(random_numbers[0] - 0.5) <= 0.5 - acceptable_error) ,iteration_count ,0)

    inner_rsqrt_det = get_ellipse_rsqrt_det(inner_ellipse)
    outer_rsqrt_det = get_ellipse_rsqrt_det(outer_ellipse)
    for i in range(0, iteration_count):
        current_dir = normalize_approx_and_flip(current_dir, quad_dirs[1])
        inner_dir = ellipse_transform(inner_ellipse, current_dir)
        outer_dir = ellipse_transform(outer_ellipse, current_dir)
        det_dirs = maximum(+0.0, dot(current_dir, rotate_90(quad_dirs[0])))
        error = target_area - get_area_between_ellipses_in_sector_from_tangents(
            inner_rsqrt_det, det_dirs / (inner_rsqrt_det * dot(quad_dirs[0], inner_dir)),
            outer_rsqrt_det, det_dirs / (outer_rsqrt_det * dot(quad_dirs[0], outer_dir)))
        quadratic = outer(inner_dir - outer_dir, rotate_90(current_dir)) - outer((2.0 * error) * inner_dir, outer_dir)
        current_dir = solve_homogeneous_quadratic(quadratic)
    current_dir = where((dot(current_dir, quad_dirs[1]) >= 0.0) ,current_dir ,-current_dir)
    inner_factor = 1.0 / get_ellipse_direction_factor_rsq(inner_ellipse, current_dir)
    outer_factor = 1.0 / get_ellipse_direction_factor_rsq(outer_ellipse, current_dir)
    current_dir *= sqrt(mix_fma(inner_factor, outer_factor, random_numbers[1]))
    return current_dir


def sample_projected_solid_angle_polygon(polygon, random_numbers):
    target_projected_solid_angle = random_numbers[0] * polygon.projected_solid_angle
    sampled_dir = zeros((3,))
    outer_ellipse = zeros((2,))
    dir_0 = zeros((2,))
    if (is_central_case(polygon)):
        for i in range(0, MAX_POLYGON_VERTEX_COUNT):
            if (i > 0):
                target_projected_solid_angle -= polygon.sector_projected_solid_angles[i - 1]
            outer_ellipse = polygon.ellipses[i]
            dir_0 = polygon.vertices[i]
            if ((i >= 2 and i + 1 == polygon.vertex_count) or target_projected_solid_angle < polygon.sector_projected_solid_angles[i]):
                break
        sqrt_det = sqrt(get_ellipse_det(outer_ellipse))
        angle = 2.0 * target_projected_solid_angle * sqrt_det
        sampled_dir[0:2] = (cos(angle) * sqrt_det) * dir_0 + sin(angle) * rotate_90(ellipse_transform(outer_ellipse, dir_0))
        sampled_dir[0:2] *= sqrt(random_numbers[1] / get_ellipse_direction_factor_rsq(outer_ellipse, sampled_dir[0:2]))
    else:
        inner_ellipse = polygon.inner_ellipse_0
        dir_1 = zeros((2,))
        for i in range(0, MAX_POLYGON_VERTEX_COUNT - 1):
            vertex_ellipse = polygon.ellipses[i]
            if (i == 0):
                outer_ellipse = vertex_ellipse
            else:
                target_projected_solid_angle -= polygon.sector_projected_solid_angles[i - 1]
                vertex_inner = is_inner_ellipse(vertex_ellipse)
                inner_ellipse = where(vertex_inner ,vertex_ellipse ,inner_ellipse)
                outer_ellipse = where(vertex_inner ,outer_ellipse ,vertex_ellipse)
            dir_0 = polygon.vertices[i]
            dir_1 = polygon.vertices[i + 1]
            sector_projected_solid_angle = polygon.sector_projected_solid_angles[i]
            if ((i >= 1 and i + 2 == polygon.vertex_count) or target_projected_solid_angle < sector_projected_solid_angle):
                break
        random_numbers[0] = target_projected_solid_angle / sector_projected_solid_angle
        sampled_dir[0:2] = sample_sector_between_ellipses(random_numbers, target_projected_solid_angle, inner_ellipse, outer_ellipse, dir_0, dir_1, 2)
    sampled_dir[2] = sqrt(maximum(0.0, fma(-sampled_dir[0], sampled_dir[0], fma(-sampled_dir[1], sampled_dir[1], 1.0))))
    return sampled_dir


def compute_projected_solid_angle_polygon_sampling_error(polygon, random_numbers, sampled_dir):
    target_projected_solid_angle = random_numbers[0] * polygon.projected_solid_angle
    if (is_central_case(polygon)):
        return vec3(0.0)
    else:
        outer_ellipse = zeros((2,))
        inner_ellipse = polygon.inner_ellipse_0
        dir_0 = zeros((2,))
        for i in range(0, MAX_POLYGON_VERTEX_COUNT - 1):
            if ((i > 1 and i + 1 == polygon.vertex_count) or (i > 0 and target_projected_solid_angle < 0.0)):
                break
            sector_projected_solid_angle = polygon.sector_projected_solid_angles[i]
            target_projected_solid_angle -= sector_projected_solid_angle
            vertex_ellipse = polygon.ellipses[i]
            vertex_inner = is_inner_ellipse(vertex_ellipse)
            if (i == 0):
                outer_ellipse = vertex_ellipse
            else:
                inner_ellipse = where(vertex_inner ,vertex_ellipse ,inner_ellipse)
                outer_ellipse = where(vertex_inner ,outer_ellipse ,vertex_ellipse)
            dir_0 = polygon.vertices[i]
        target_projected_solid_angle += sector_projected_solid_angle
        sampled_projected_solid_angle = get_area_between_ellipses_in_sector(
            inner_ellipse, get_ellipse_rsqrt_det(inner_ellipse),
            outer_ellipse, get_ellipse_rsqrt_det(outer_ellipse),
            dir_0, sampled_dir[0:2])
        scaled_backward_error = target_projected_solid_angle - sampled_projected_solid_angle
        backward_error = scaled_backward_error / polygon.projected_solid_angle
        constraint_matrix = mat2()
        inner_dir = ellipse_transform(inner_ellipse, sampled_dir[0:2])
        outer_dir = ellipse_transform(outer_ellipse, sampled_dir[0:2])
        inner_factor = 1.0 / dot(sampled_dir[0:2], inner_dir)
        outer_factor = 1.0 / dot(sampled_dir[0:2], outer_dir)
        constraint_matrix[0] = 0.5 * (inner_factor - outer_factor) * rotate_90(sampled_dir[0:2])
        constraint_matrix[1] = ((1.0 - random_numbers[1]) / (inner_factor * inner_factor)) * inner_dir
        constraint_matrix[1] += (random_numbers[1] / (outer_factor * outer_factor)) * outer_dir
        constraint_matrix = transpose(constraint_matrix)
        sample_derivative = zeros((3,))
        sample_derivative[0:2] = (1.0 / linalg.det(constraint_matrix)) * vec2(constraint_matrix[1][1], -constraint_matrix[1][0])
        sample_derivative[2] = -dot(sampled_dir[0:2], sample_derivative[0:2]) / sampled_dir[2]
        forward_error = length(sample_derivative) * scaled_backward_error
        return vec3(backward_error, scaled_backward_error, forward_error)
