#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from matplotlib import pyplot, rcParams, colors
from matplotlib.backend_bases import MouseButton
from imageio import imread
import time
from os import path, makedirs
from subprocess import Popen


class Presentation:
    """A collection of slides and information about their common layout."""

    def __init__(self):
        """Initializes with no slides and for 1920x1080 presentation."""
        # The slide size in centimeters/data units
        self.size = (16.0, 9.0)
        self.dots_per_centimeter = 1920.0 / self.size[0]
        self.slides = list()

    def create_figure(self):
        """
        Creates and returns a suitable matplotlib figure for displaying this
        presentation.
        """
        rcParams["toolbar"] = "None"
        return pyplot.figure(
            frameon=False, tight_layout=dict(pad=0.0, w_pad=0.0, h_pad=0.0),
            figsize=np.asarray(self.size) / 2.54, dpi=self.dots_per_centimeter * 2.54)

    def create_axes(self):
        """
        Creates axes in the current figure, suitable for drawing to the whole
        area of the figure created by self.create_figure().
        """
        return pyplot.subplot(aspect=1.0, frameon=False, xticks=[], yticks=[], xlim=(0.0, self.size[0]), ylim=(0.0, self.size[1]))

    def rectify_state(self, slide, step):
        """
        Given a slide and step index, this function returns a pair that is in
        range for this presentation. In- or decrementing the step in the given
        state always corresponds to one step in the output, except at
        presentation boundaries, where indices get clamped. Valid inputs remain
        unchanged.
        """
        step_counts = [0] + [slide.step_count for slide in self.slides]
        summed_counts = np.cumsum(step_counts)
        # Apply clamping
        if slide < 0:
            slide, step = 0, 0
        if slide >= len(self.slides):
            slide, step = len(self.slides) - 1, (step_counts[-1] - 1)
        # Ravel, clamp and unravel the index
        flat_index = summed_counts[slide] + step
        flat_index = np.clip(flat_index, 0, summed_counts[-1] - 1)
        slide = np.argmax(summed_counts > flat_index) - 1
        return slide, flat_index - summed_counts[slide]

    def draw_slide(self, slide, step):
        """Draws a slide including overlays to the current axes."""
        slide, step = self.rectify_state(slide, step)
        current_slide = self.slides[slide]
        current_slide.step = step
        current_slide.step_lerp = 0.0
        axes = pyplot.gca()
        current_slide.draw()
        pyplot.sca(axes)
        self.draw_overlay(slide, step)

    def draw_overlay(self, slide, step):
        """Draws an overlay (e.g. slide numbers) to the active axes."""
        pass


class PresentationWithFooter(Presentation):
    """
    A presentation with slide numbers at the right bottom and a logo at the left
    bottom.
    """

    def __init__(self, logo_path, logo_width, no_footer_slides=()):
        """
        :param logo_path: Path to the image file for the logo.
        :param logo_width: Width of the logo in data units.
        :param no_footer_slides: Iterable of slide indices where the footer is
            hidden.
        """
        super().__init__()
        self.logo = imread(logo_path)
        self.logo_width = logo_width
        self.no_footer_slides = frozenset(no_footer_slides)

    def draw_overlay(self, slide, step):
        if slide not in self.no_footer_slides:
            draw_image(self.logo, (0.15, 0.1), self.logo_width)
            pyplot.text(self.size[0] - 0.15, 0.15, str(slide), ha="right", fontsize=8.0, color=3 * [0.7])


class Slideshow:
    """Handles the figure for a slideshow."""

    def __init__(self, presentation, slide=0, step=0):
        """Creates the figure."""
        self.presentation = presentation
        self.displayed_state = -1, -1
        self.state = self.presentation.rectify_state(slide, step)
        self.previous_mouse_xy = (None, None)
        self.figure = self.presentation.create_figure()
        self.timer = None
        self.held_buttons = frozenset()
        # time.monotonic() for the beginning and end of a step transition or
        # 0.0, 0.0 if no transition is in progress
        self.step_begin, self.step_end = 0.0, 0.0

    def connect_handlers(self):
        """
        Defines and connects a bunch of event handlers to the figure.
        """
        def key_press_event(event):
            if event.key == "left":
                self.step_backward()
                self.update()
            elif event.key == "right":
                self.step_forward()
                self.update()
            elif event.key == "up":
                self.slide_backward()
            elif event.key == "down":
                self.slide_forward()
            else:
                self.set_slide_attributes()
                self.update(self.current_slide().on_key(event.key))
        self.figure.canvas.mpl_connect("key_press_event", key_press_event)

        button_dict = {MouseButton.LEFT: "lmb", MouseButton.MIDDLE: "mmb", MouseButton.RIGHT: "rmb"}

        def button_press_event(event):
            if event.button in button_dict:
                self.held_buttons = self.held_buttons | {button_dict[event.button]}
                self.update(self.current_slide().on_click((event.x, event.y), button_dict[event.button]))
        self.figure.canvas.mpl_connect("button_press_event", button_press_event)

        def button_release_event(event):
            if event.button in button_dict:
                self.held_buttons = self.held_buttons - {button_dict[event.button]}
        self.figure.canvas.mpl_connect("button_release_event", button_release_event)

        def figure_leave_event(event):
            # Just to be safe we assume that all buttons are released while the
            # mouse is outside the figure
            self.held_buttons = frozenset()
        self.figure.canvas.mpl_connect("figure_leave_event", figure_leave_event)

        def motion_notify_event(event):
            if self.previous_mouse_xy[0] is None:
                self.previous_mouse_xy = event.x, event.y
            motion = event.x - self.previous_mouse_xy[0], event.y - self.previous_mouse_xy[1]
            self.set_slide_attributes()
            self.set_slide_attributes()
            self.update(self.current_slide().on_mouse_move((event.x, event.y), motion, self.held_buttons))
            self.previous_mouse_xy = event.x, event.y
        self.figure.canvas.mpl_connect("motion_notify_event", motion_notify_event)

        def timer_event():
            slide = self.current_slide()
            if self.displayed_state == self.state:
                now = self.set_slide_attributes()
                if self.step_begin != self.step_end:
                    if now < self.step_begin or now < self.step_end:
                        self.update(slide.animate())
                    elif self.step_begin < self.step_end:
                        self.step_forward(True)
                    elif self.step_end < self.step_begin:
                        self.update(slide.animate())
                        self.step_begin = self.step_end = 0.0
                else:
                    self.update(slide.animate())
            self.timer.interval = 1
        self.timer = self.figure.canvas.new_timer(1, [(timer_event, tuple(), dict())])

    def start(self):
        """Starts displaying the slideshow."""
        self.update()
        self.connect_handlers()
        self.timer.start()
        pyplot.show()

    def redraw_canvas(self):
        """Draws the canvas of the figure such that changes become visible."""
        self.figure.canvas.draw()

    def step_forward(self, force=False):
        """
        Initializes the next step, taking the step time specified by the slide
        into account.
        :param force: True to complete the step instantaneously.
        """
        # If a forward step is already in progress, we force it
        if not force and self.step_begin < self.step_end:
            return self.step_forward(True)
        step_time = self.current_slide().step_time(self.state[1])
        if force or step_time == 0.0:
            self.state = self.presentation.rectify_state(self.state[0], self.state[1] + 1)
            self.step_begin = self.step_end = 0.0
            self.update()
        else:
            self.step_begin = time.monotonic()
            self.step_end = self.step_begin + step_time

    def step_backward(self, force=False):
        """
        Initializes a step backwards, taking the step time specified by the
        slide into account.
        :param force: True to complete the step instantaneously.
        """
        # If a forward step is in progress, we just abort it
        if self.step_begin < self.step_end:
            self.step_begin, self.step_end = 0, 0
            return
        step_time = self.current_slide().step_time(self.state[1] - 1)
        self.state = self.presentation.rectify_state(self.state[0], self.state[1] - 1)
        self.update()
        if force or step_time == 0.0:
            self.step_begin = self.step_end = 0.0
        else:
            self.step_end = time.monotonic()
            self.step_begin = self.step_end + step_time

    def slide_forward(self):
        """Forces going forward to the next slide (if any) immediately."""
        self.state = self.presentation.rectify_state(self.state[0] + 1, 0)
        self.step_begin = self.step_end = 0.0
        self.update()

    def slide_backward(self, force=False):
        """Forces going back to the previous slide (if any) immediately."""
        self.state = self.presentation.rectify_state(self.state[0] - 1, 0)
        self.step_begin = self.step_end = 0.0
        self.update()

    def current_slide(self):
        """
        :return: The Slide that should be displayed currently (but may not be
            displayed until after the next update()).
        """
        return self.presentation.slides[self.state[0]]

    def set_slide_attributes(self):
        """
        Ensures that attributes for self.current_slide() like the current step
        or the lerp factor are up to date.
        :return: The value of time.monotonic() used to set time-dependent
            attributes.
        """
        slide = self.current_slide()
        slide.step = self.state[1]
        slide.dots_per_centimeter = self.presentation.dots_per_centimeter
        now = time.monotonic()
        if self.step_begin != self.step_end:
            slide.step_lerp = (now - self.step_begin) / (self.step_end - self.step_begin)
            slide.step_lerp = np.clip(slide.step_lerp, 0.0, 1.0)
            slide.lerp_ongoing = True
        else:
            slide.step_lerp = 0.0
            slide.lerp_ongoing = False
        return now

    def draw(self, redraw_canvas=True):
        """
        Invokes all the commands necessary to draw this slide, including
        overlays. Optionally it also redraws the canvas.
        """
        self.figure.clear()
        axes = self.presentation.create_axes()
        slide = self.current_slide()
        self.set_slide_attributes()
        slide.draw()
        pyplot.sca(axes)
        self.presentation.draw_overlay(self.state[0], self.state[1])
        if redraw_canvas:
            self.redraw_canvas()

    def on_step(self, state, previous_state):
        """Forwards to slide.on_step() and displays notes."""
        slide = self.current_slide()
        if len(slide.notes) > state[1]:
            show_notes(slide.notes[state[1]], state[0] != previous_state[0])
        self.update(slide.on_step(state, previous_state))

    def update(self, command=None):
        """
        Ensures that the current slide is also the displayed slide.
        Additionally, it can respond to several commands from Slide callbacks.
        :param command: This can be one of the following strings, which invokes
            the matching method of this class: redraw_canvas, step_forward,
            step_backward, draw.
        """
        if command == "redraw_canvas":
            self.redraw_canvas()
        elif command == "step_forward":
            self.step_forward()
        elif command == "step_backward":
            self.step_backward()
        elif command == "draw":
            self.draw()
        elif self.displayed_state != self.state:
            old_state = tuple(self.displayed_state)
            self.displayed_state = self.state
            self.draw(False)
            slide = self.current_slide()
            self.on_step(self.state, old_state)
            self.redraw_canvas()


class Slide:
    """Base class for slides."""

    def __init__(self):
        """Initializes as single step slide without animated transitions."""
        self.notes = []
        self.step_count = 1
        # The step that was most recently displayed by a slide show. Updated
        # before any of the callbacks are invoked.
        self.step = -1
        # Transition times from each step to the next in seconds. Implicitly
        # padded with zeros.
        self.step_durations = [0.0]
        # The linear interpolation factor (i.e. a weight between zero and one
        # that changes with constant velocity) for the transition from the
        # current step to the next.
        self.step_lerp = 0.0
        # Whether a step transition with lerp is ongoing
        self.lerp_ongoing = False
        # The factor used to convert from data units to pixels
        self.dots_per_centimeter = -1.0

    def draw(self):
        """
        Draws the current step of this slide to the (cleared) current axes.
        """
        pass

    def step_time(self, step):
        """:return: The step time for the step with the given index."""
        return self.step_durations[step] if 0 <= step < len(self.step_durations) else 0.0

    def animate(self):
        """
        Updates drawings produced by draw() to implement an animation. Called in
        regular intervals.
        :return: A command for Slideshow.update() (optional).
        """
        pass

    def on_key(self, key):
        """
        Invoked on a key press, unless that key press corresponds to global
        slide controls.
        :param key: A string describing the key. See
            matplotlib.backend_bases.KeyEvent.key.
        :return: A command for Slideshow.update() (optional).
        """
        pass

    def on_click(self, position, button):
        """
        Invoked on a mouse click (i.e. when the button was pressed down just now) in the figure.
        :param position: Offset from the left bottom of the canvas in pixels.
        :param button: Either "lmb", "mmb" or "rmb".
        :return: A command for Slideshow.update() (optional).
        """
        pass

    def on_mouse_move(self, position, motion, held_buttons):
        """
        Invoked when the mouse has moved across the figure.
        :param position: Offset from the left bottom of the canvas in pixels.
        :param motion: Change of the mouse position since the last invocation in
            pixels.
        :param held_buttons: A frozenset of strings "lmb", "mmb" or "rmb". It
            gives all mouse buttons that are currently down.
        :return: A command for Slideshow.update() (optional).
        """
        pass

    def on_step(self, state, previous_state):
        """
        Invoked when the presentation starts displaying this slide or a new step
        on this slide, just after the corresponding draw() call.
        :param state: Slide and step index that are being displayed now.
        :param previous_state: Slide and step index that were displayed before
            (both -1 if the presentation just began).
        """
        pass


class FunctionSlide(Slide):
    """A non-interactive slide defined conveniently by a single function."""

    def __init__(self, function, step_count=1):
        """
        Uses the given function to draw the slide. The function gets the step
        index as parameter and cannot return anything. The step count is passed
        explicitly.
        """
        super().__init__()
        self.function = function
        self.step_count = step_count

    def draw(self):
        """Forwards to the function."""
        self.function(self.step)


class ImageComparisonSlide(Slide):
    """A slide showing multiple images with a synchronized magnifier."""

    def __init__(self):
        """Initializes with no images to compare."""
        super().__init__()
        # List of 3D arrays for images to be displayed
        self.images = list()
        # List holding one label string for each image
        self.labels = list()
        # Left bottom data coordinate for the first image
        self.left_bottom = (1.0, 1.0)
        # Width in data coordinates for all images
        self.image_width = 1.0
        # Offset in data coordinates along x and y from one image to the next
        self.offset = (1.0, 0.0)
        # The zoom factor for the magnified inset
        self.zoom = 8
        # The half width and height of the magnified region in pixels
        self.half_magnifier_extent = (16, 16)
        # The width of the outline around the magnifier in pixels
        self.border_width = 10
        # The color of the magnifier outline (any matplotlib color specification)
        self.border_color = "C1"
        # The current center location of the magnifier in pixels within the
        # first image or (-1, -1) if there is no magnifier
        self.magnifier_center = -1, -1

    def on_click(self, position, button):
        """Places the magnifier or removes it."""
        self.magnifier_center = -1, -1
        x, y = np.asarray(position) / self.dots_per_centimeter
        for i, image in enumerate(self.images):
            left, bottom = self.left_bottom[0] + i * self.offset[0], self.left_bottom[1] + i * self.offset[1]
            width, height = self.image_width, self.image_width * image.shape[0] / image.shape[1]
            if left <= x < left + width and bottom <= y < bottom + height:
                self.magnifier_center = (
                    np.minimum(int(image.shape[1] * (x - left) / width), image.shape[1] - 1),
                    np.minimum(int(image.shape[0] * (bottom + height - y) / height), image.shape[0] - 1),
                )
                break
        return "draw"

    def draw(self):
        """Draws all images including magnifiers and labels."""
        border_color = np.asarray(np.asarray(colors.to_rgb(self.border_color)) * 255.0, dtype=np.uint8)
        for i, image in enumerate(self.images):
            label = self.labels[i] if len(self.labels) > i else ""
            # Apply the magnifier
            if self.magnifier_center[0] >= 0:
                center_x, center_y = self.magnifier_center
                center_x = np.clip(center_x, self.half_magnifier_extent[0] * self.zoom + self.border_width, image.shape[1] - self.half_magnifier_extent[0] * self.zoom - self.border_width)
                center_y = np.clip(center_y, self.half_magnifier_extent[1] * self.zoom + self.border_width, image.shape[0] - self.half_magnifier_extent[1] * self.zoom - self.border_width)
                left, right = center_x - self.half_magnifier_extent[0] * self.zoom, center_x + self.half_magnifier_extent[0] * self.zoom
                top, bottom = center_y - self.half_magnifier_extent[1] * self.zoom, center_y + self.half_magnifier_extent[1] * self.zoom
                border_left, border_right = left - self.border_width, right + self.border_width
                border_top, border_bottom = top - self.border_width, bottom + self.border_width
                crop = np.copy(image[center_y - self.half_magnifier_extent[1]:center_y + self.half_magnifier_extent[1], center_x - self.half_magnifier_extent[0]:center_x + self.half_magnifier_extent[0]])
                image = np.copy(image)
                image[border_top:border_bottom, border_left:border_right] = border_color
                image[top:bottom, left:right] = crop.repeat(self.zoom, axis=0).repeat(self.zoom, axis=1)
            # Draw the image
            left_bottom = self.left_bottom[0] + i * self.offset[0], self.left_bottom[1] + i * self.offset[1]
            draw_image(image, left_bottom, self.image_width)
            # Draw the label
            if len(label) > 0:
                center = self.left_bottom[0] + i * self.offset[0] + self.image_width * 0.5
                top = self.left_bottom[1] + self.image_width * image.shape[0] / image.shape[1]
                pyplot.text(center, top + 0.05, label, ha="center", va="bottom")


def show_notes(notes, new_slide=True):
    """
    Makes notes for a slide or a step visible at a suitable location (currently
    stdout).
    """
    if new_slide:
        print()
        print()
    print(notes)


def lerp(begin, end, factor):
    """Blends between begin and end linearly."""
    factor = np.clip(factor, 0.0, 1.0)
    return begin * (1.0 - factor) + end * factor


def smooth_lerp(begin, end, factor):
    """Blends between begin and end smoothly using a cubic."""
    factor = np.clip(factor, 0.0, 1.0)
    factor = (3.0 - 2.0 * factor) * factor * factor
    return lerp(begin, end, factor)


def draw_title(text):
    """Plots a title with appropriate font and position."""
    pyplot.text(1.0, 8.0, text, fontsize=18.0)


def draw_bullets(texts, line_height=1.0, top_y=7.0):
    """
    This function plots several lines of text.
    :param texts: Iterable of strings, one per line.
    :param line_height: Vertical offset between two subsequent lines in data
        units.
    :param top_y: y-coordinate in data units for the line at the top.
    :return: The x and y-coordinate that would be used by the next bullet.
    """
    y = top_y
    for i, text_or_space in enumerate(texts):
        if isinstance(text_or_space, str):
            text = text_or_space
            pyplot.text(1.0, y, text)
            y -= line_height
        else:
            y -= text_or_space
    return 1.0, y


def draw_image(path_or_array, left_bottom, width, height=None):
    """
    Creates axes for an image within the current axes and uses them to display
    an image.
    :param path_or_array: Either a path to an image (i.e. a string) or a NumPy
        array of an image as expected by pyplot.imshow().
    :param left_bottom: The left bottom of the image in data coordinates of the
        current axes.
    :param width: The width of the image in data coordinates.
    :param height: Height in data coordinates. Pass None to preserve the aspect
        ratio.
    :return: The newly created axes.
    """
    if isinstance(path_or_array, str):
        image = imread(path_or_array)
    else:
        image = path_or_array
    if height is None:
        height = width * image.shape[0] / image.shape[1]
    rectangle = left_bottom[0], left_bottom[1], width, height
    axes = pyplot.gca().inset_axes(rectangle, transform=pyplot.gca().transData, xticks=[], yticks=[], frameon=False)
    axes.imshow(image)


def export_presentation_pdf(presentation, output_path, slide_steps=True, pdftk_path="pdftk", ghostscript_path="gs"):
    """
    Exports the given presentation (in its entirety) as PDF. Some intermediate
    files are created in a directory next to the output file.
    :param presentation: A Presentation object to export.
    :param output_path: The path to the PDF file that should be created.
    :param slide_steps: True to export one page for the last step on each slide.
        False to export one page per step.
    :param pdftk_path: Path to the binary for pdftk (used to concatenate PDFs).
        Pass None to use ghostscript.
    :param ghostscript_path: Path to the binary for ghostscript.
    """
    # List the pages
    step_counts = [slide.step_count for slide in presentation.slides]
    if slide_steps:
        pages = enumerate(np.asarray(step_counts) - 1)
    else:
        pages = [(slide, step) for slide, step_count in enumerate(step_counts) for step in range(step_count)]
    # Prepare an intermediate directory
    temp_dir = path.splitext(output_path)[0] + "_temp"
    if not path.exists(temp_dir):
        makedirs(temp_dir)
    # Create the figure and draw the pages one by one
    figure = presentation.create_figure()
    file_list = list()
    for i, state in enumerate(pages):
        figure.clear()
        presentation.create_axes()
        presentation.draw_slide(state[0], state[1])
        file = path.join(temp_dir, "%04d.pdf" % i)
        figure.savefig(file)
        file_list.append(file)
    # Concatenate the PDFs
    if pdftk_path is not None:
        Popen([pdftk_path] + file_list + ["cat", "output", output_path])
    else:
        Popen([ghostscript_path, "-dBATCH", "-dNOPAUSE", "-q", "-sDEVICE=pdfwrite",
               "-dPDFSETTINGS=/prepress", "-sOutputFile=%s" % output_path] + file_list)
